"""
brozzler/dashboard/__init__.py - flask app for brozzler dashboard, defines api
endspoints etc

Copyright (C) 2014-2017 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import base64
import os
import sys

import doublethink
import rethinkdb as rdb
import structlog
import yaml

logger = structlog.get_logger(logger_name=__name__)

try:
    import flask
except ImportError as e:
    logger.critical(
        '%s: %s\n\nYou might need to run "pip install '
        'brozzler[dashboard]".\nSee README.rst for more information.',
        type(e).__name__,
        e,
    )
    sys.exit(1)

r = rdb.RethinkDB()

app = flask.Flask(__name__)

# configure with environment variables
SETTINGS = {
    "RETHINKDB_SERVERS": os.environ.get(
        "BROZZLER_RETHINKDB_SERVERS", "localhost"
    ).split(","),
    "RETHINKDB_DB": os.environ.get("BROZZLER_RETHINKDB_DB", "brozzler"),
    "WAYBACK_BASEURL": os.environ.get(
        "WAYBACK_BASEURL", "http://localhost:8880/brozzler"
    ),
    "DASHBOARD_PORT": os.environ.get("DASHBOARD_PORT", "8000"),
    "DASHBOARD_INTERFACE": os.environ.get("DASHBOARD_INTERFACE", "localhost"),
}
rr = doublethink.Rethinker(SETTINGS["RETHINKDB_SERVERS"], db=SETTINGS["RETHINKDB_DB"])
_svc_reg = None


def service_registry():
    global _svc_reg
    if not _svc_reg:
        _svc_reg = doublethink.ServiceRegistry(rr)
    return _svc_reg


@app.route("/api/sites/<site_id>/queued_count")
@app.route("/api/site/<site_id>/queued_count")
def queued_count(site_id):
    reql = (
        rr.table("pages")
        .between(
            [site_id, 0, False, r.minval],
            [site_id, 0, False, r.maxval],
            index="priority_by_site",
        )
        .count()
    )
    logger.debug("querying rethinkdb", query=reql)
    count = reql.run()
    return flask.jsonify(count=count)


@app.route("/api/sites/<site_id>/queue")
@app.route("/api/site/<site_id>/queue")
def queue(site_id):
    logger.debug("flask.request.args", args=flask.request.args)
    start = flask.request.args.get("start", 0)
    end = flask.request.args.get("end", start + 90)
    reql = rr.table("pages").between(
        [site_id, 0, False, r.minval],
        [site_id, 0, False, r.maxval],
        index="priority_by_site",
    )[start:end]
    logger.debug("querying rethinkdb", query=reql)
    queue_ = reql.run()
    return flask.jsonify(queue_=list(queue_))


@app.route("/api/sites/<site_id>/pages_count")
@app.route("/api/site/<site_id>/pages_count")
@app.route("/api/sites/<site_id>/page_count")
@app.route("/api/site/<site_id>/page_count")
def page_count(site_id):
    reql = (
        rr.table("pages")
        .between(
            [site_id, 1, False, r.minval],
            [site_id, r.maxval, False, r.maxval],
            index="priority_by_site",
        )
        .count()
    )
    logger.debug("querying rethinkdb", query=reql)
    count = reql.run()
    return flask.jsonify(count=count)


@app.route("/api/sites/<site_id>/pages")
@app.route("/api/site/<site_id>/pages")
def pages(site_id):
    """Pages already crawled."""
    start = int(flask.request.args.get("start", 0))
    end = int(flask.request.args.get("end", start + 90))
    reql = (
        rr.table("pages")
        .between(
            [site_id, 1, r.minval], [site_id, r.maxval, r.maxval], index="least_hops"
        )
        .order_by(index="least_hops")[start:end]
    )
    logger.debug("querying rethinkdb", query=reql)
    pages_ = reql.run()
    return flask.jsonify(pages=list(pages_))


@app.route("/api/pages/<page_id>")
@app.route("/api/page/<page_id>")
def page(page_id):
    reql = rr.table("pages").get(page_id)
    logger.debug("querying rethinkdb", query=reql)
    page_ = reql.run()
    return flask.jsonify(page_)


@app.route("/api/pages/<page_id>/yaml")
@app.route("/api/page/<page_id>/yaml")
def page_yaml(page_id):
    reql = rr.table("pages").get(page_id)
    logger.debug("querying rethinkdb", query=reql)
    page_ = reql.run()
    return app.response_class(
        yaml.dump(page_, default_flow_style=False), mimetype="application/yaml"
    )


@app.route("/api/sites/<site_id>")
@app.route("/api/site/<site_id>")
def site(site_id):
    reql = rr.table("sites").get(site_id)
    logger.debug("querying rethinkdb", query=reql)
    s = reql.run()
    if "cookie_db" in s:
        s["cookie_db"] = base64.b64encode(s["cookie_db"]).decode("ascii")
    return flask.jsonify(s)


@app.route("/api/sites/<site_id>/yaml")
@app.route("/api/site/<site_id>/yaml")
def site_yaml(site_id):
    reql = rr.table("sites").get(site_id)
    logger.debug("querying rethinkdb", query=reql)
    site_ = reql.run()
    return app.response_class(
        yaml.dump(site_, default_flow_style=False), mimetype="application/yaml"
    )


@app.route("/api/stats/<bucket>")
def stats(bucket):
    reql = rr.table("stats").get(bucket)
    logger.debug("querying rethinkdb", query=reql)
    stats_ = reql.run()
    return flask.jsonify(stats_)


@app.route("/api/jobs/<job_id>/sites")
@app.route("/api/job/<job_id>/sites")
def sites(job_id):
    try:
        jid = int(job_id)
    except ValueError:
        jid = job_id
    reql = rr.table("sites").get_all(jid, index="job_id")
    logger.debug("querying rethinkdb", query=reql)
    sites_ = list(reql.run())
    # TypeError: <binary, 7168 bytes, '53 51 4c 69 74 65...'> is not JSON serializable
    for s in sites_:
        if "cookie_db" in s:
            s["cookie_db"] = base64.b64encode(s["cookie_db"]).decode("ascii")
    return flask.jsonify(sites=sites_)


@app.route("/api/jobless-sites")
def jobless_sites():
    # XXX inefficient (unindexed) query
    reql = rr.table("sites").filter(~r.row.has_fields("job_id"))
    logger.debug("querying rethinkdb", query=reql)
    sites_ = list(reql.run())
    # TypeError: <binary, 7168 bytes, '53 51 4c 69 74 65...'> is not JSON serializable
    for s in sites_:
        if "cookie_db" in s:
            s["cookie_db"] = base64.b64encode(s["cookie_db"]).decode("ascii")
    return flask.jsonify(sites=sites_)


@app.route("/api/jobs/<job_id>")
@app.route("/api/job/<job_id>")
def job(job_id):
    try:
        jid = int(job_id)
    except ValueError:
        jid = job_id
    reql = rr.table("jobs").get(jid)
    logger.debug("querying rethinkdb", query=reql)
    job_ = reql.run()
    return flask.jsonify(job_)


@app.route("/api/jobs/<job_id>/yaml")
@app.route("/api/job/<job_id>/yaml")
def job_yaml(job_id):
    try:
        jid = int(job_id)
    except ValueError:
        jid = job_id
    reql = rr.table("jobs").get(jid)
    logger.debug("querying rethinkdb", query=reql)
    job_ = reql.run()
    return app.response_class(
        yaml.dump(job_, default_flow_style=False), mimetype="application/yaml"
    )


@app.route("/api/workers")
def workers():
    workers_ = service_registry().available_services("brozzler-worker")
    return flask.jsonify(workers=list(workers_))


@app.route("/api/services")
def services():
    services_ = service_registry().available_services()
    return flask.jsonify(services=list(services_))


@app.route("/api/jobs")
def jobs():
    reql = rr.table("jobs").order_by(r.desc("id"))
    logger.debug("querying rethinkdb", query=reql)
    jobs_ = list(reql.run())
    return flask.jsonify(jobs=jobs_)


@app.route("/api/config")
def config():
    return flask.jsonify(config=SETTINGS)


@app.route("/api/<path:path>")
@app.route("/api", defaults={"path": ""})
def api404(path):
    flask.abort(404)


@app.route("/", defaults={"path": ""})
@app.route("/<path:path>")
def root(path):
    return flask.render_template("index.html")


try:
    import logging

    import gunicorn.app.base
    import gunicorn.glogging
    from gunicorn.six import iteritems

    class BypassGunicornLogging(gunicorn.glogging.Logger):
        def setup(self, cfg):
            self.error_log.handlers = logging.root.handlers
            self.access_log.handlers = logging.root.handlers

    class GunicornBrozzlerDashboard(gunicorn.app.base.BaseApplication):
        def __init__(self, app, options=None):
            self.options = options or {}
            self.application = app
            super(GunicornBrozzlerDashboard, self).__init__()

        def load_config(self):
            config = dict(
                [
                    (key, value)
                    for key, value in iteritems(self.options)
                    if key in self.cfg.settings and value is not None
                ]
            )
            for key, value in iteritems(config):
                self.cfg.set(key.lower(), value)
            self.cfg.set("logger_class", BypassGunicornLogging)
            self.cfg.set("accesslog", "dummy-value")

        def load(self):
            return self.application

    def run(**options):
        logger.info("running brozzler-dashboard using gunicorn")
        GunicornBrozzlerDashboard(app, options).run()

except ImportError:

    def run():
        logger.info("running brozzler-dashboard using simple flask app.run")
        app.run(host=SETTINGS["DASHBOARD_INTERFACE"], port=SETTINGS["DASHBOARD_PORT"])


def main(argv=None):
    import argparse

    import brozzler.cli

    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(
            "brozzler-dashboard - web application for viewing brozzler crawl status"
        ),
        epilog=(
            "brozzler-dashboard has no command line options, but can be "
            "configured using the following environment variables:\n\n"
            "  BROZZLER_RETHINKDB_SERVERS   rethinkdb servers, e.g. "
            "db0.foo.org,db0.foo.org:38015,db1.foo.org (default: "
            "localhost)\n"
            "  BROZZLER_RETHINKDB_DB        rethinkdb database name "
            "(default: brozzler)\n"
            "  WAYBACK_BASEURL     base url for constructing wayback "
            "links (default http://localhost:8880/brozzler)"
            "  DASHBOARD_PORT   brozzler-dashboard listening port (default: 8000)\n"
            "  DASHBOARD_INTERFACE brozzler-dashboard network interface binding (default: localhost)"
        ),
    )
    brozzler.cli.add_common_options(arg_parser, argv)
    args = arg_parser.parse_args(args=argv[1:])
    brozzler.cli.configure_logging(args)
    run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
brozzler-easy - brozzler-worker, warcprox, pywb, and brozzler-dashboard all
working together in a single process

Copyright (C) 2016-2018 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import argparse
import os
import signal
import socket
import socketserver
import sys
import threading
import time
import traceback

import doublethink
import structlog

import brozzler
import brozzler.cli

logger = structlog.get_logger(logger_name=__name__)

try:
    import wsgiref.handlers
    import wsgiref.simple_server

    import pywb
    import warcprox
    import warcprox.main

    import brozzler.dashboard
    import brozzler.pywb
except ImportError as e:
    logger.critical(
        '%s: %s\n\nYou might need to run "pip install '
        'brozzler[easy]".\nSee README.rst for more information.',
        type(e).__name__,
        exc_info=True,
    )
    sys.exit(1)


def _build_arg_parser(argv=None):
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        formatter_class=brozzler.cli.BetterArgumentDefaultsHelpFormatter,
        prog=os.path.basename(argv[0]),
        description=(
            "brozzler-easy - easy deployment of brozzler, with "
            "brozzler-worker, warcprox, pywb, and brozzler-dashboard all "
            "running in a single process"
        ),
    )

    # common args
    brozzler.cli.add_rethinkdb_options(arg_parser)
    arg_parser.add_argument(
        "-d",
        "--warcs-dir",
        dest="warcs_dir",
        default="./warcs",
        help="where to write warcs",
    )

    # warcprox args
    arg_parser.add_argument(
        "-c",
        "--cacert",
        dest="cacert",
        default="./%s-warcprox-ca.pem" % socket.gethostname(),
        help=(
            "warcprox CA certificate file; if file does not exist, it will be created"
        ),
    )
    arg_parser.add_argument(
        "--certs-dir",
        dest="certs_dir",
        default="./%s-warcprox-ca" % socket.gethostname(),
        help="where warcprox will store and load generated certificates",
    )
    arg_parser.add_argument(
        "--onion-tor-socks-proxy",
        dest="onion_tor_socks_proxy",
        default=None,
        help=("host:port of tor socks proxy, used only to connect to .onion sites"),
    )

    # brozzler-worker args
    arg_parser.add_argument(
        "-e",
        "--chrome-exe",
        dest="chrome_exe",
        default=brozzler.cli.suggest_default_chrome_exe(),
        help="executable to use to invoke chrome",
    )
    arg_parser.add_argument(
        "-n",
        "--max-browsers",
        dest="max_browsers",
        type=int,
        default=1,
        help=("max number of chrome instances simultaneously browsing pages"),
    )

    # pywb args
    arg_parser.add_argument(
        "--pywb-address",
        dest="pywb_address",
        default="0.0.0.0",
        help="pywb wayback address to listen on",
    )
    arg_parser.add_argument(
        "--pywb-port",
        dest="pywb_port",
        type=int,
        default=8880,
        help="pywb wayback port",
    )

    # dashboard args
    arg_parser.add_argument(
        "--dashboard-address",
        dest="dashboard_address",
        default="localhost",
        help="brozzler dashboard address to listen on",
    )
    arg_parser.add_argument(
        "--dashboard-port",
        dest="dashboard_port",
        type=int,
        default=8881,
        help="brozzler dashboard port",
    )

    # common at the bottom args
    brozzler.cli.add_common_options(arg_parser, argv)

    return arg_parser


class ThreadingWSGIServer(
    socketserver.ThreadingMixIn, wsgiref.simple_server.WSGIServer
):
    pass


class BrozzlerEasyController:
    logger = structlog.get_logger(logger_name=__module__ + "." + __qualname__)

    def __init__(self, args):
        self.stop = threading.Event()
        self.args = args
        self.warcprox_controller = warcprox.controller.WarcproxController(
            self._warcprox_opts(args)
        )
        self.brozzler_worker = self._init_brozzler_worker(args)
        self.pywb_httpd = self._init_pywb(args)
        self.dashboard_httpd = self._init_brozzler_dashboard(args)

    def _init_brozzler_dashboard(self, args):
        return wsgiref.simple_server.make_server(
            args.dashboard_address,
            args.dashboard_port,
            brozzler.dashboard.app,
            ThreadingWSGIServer,
        )

    def _init_brozzler_worker(self, args):
        rr = doublethink.Rethinker(args.rethinkdb_servers.split(","), args.rethinkdb_db)
        frontier = brozzler.RethinkDbFrontier(rr)
        service_registry = doublethink.ServiceRegistry(rr)
        worker = brozzler.worker.BrozzlerWorker(
            frontier,
            service_registry,
            chrome_exe=args.chrome_exe,
            proxy="%s:%s" % self.warcprox_controller.proxy.server_address,
            max_browsers=args.max_browsers,
        )
        return worker

    def _init_pywb(self, args):
        brozzler.pywb.TheGoodUrlCanonicalizer.replace_default_canonicalizer()
        brozzler.pywb.TheGoodUrlCanonicalizer.monkey_patch_dsrules_init()
        brozzler.pywb.support_in_progress_warcs()
        brozzler.pywb.monkey_patch_wburl()
        brozzler.pywb.monkey_patch_fuzzy_query()
        brozzler.pywb.monkey_patch_calc_search_range()

        if args.warcs_dir.endswith("/"):
            warcs_dir = args.warcs_dir
        else:
            warcs_dir = args.warcs_dir + "/"

        conf = {
            "collections": {
                "brozzler": {
                    "index_paths": brozzler.pywb.RethinkCDXSource(
                        servers=args.rethinkdb_servers.split(","),
                        db=args.rethinkdb_db,
                        table="captures",
                    )
                },
            },
            # 'enable_http_proxy': True,
            # 'enable_memento': True,
            "archive_paths": warcs_dir,
            "enable_cdx_api": True,
            "framed_replay": True,
            "port": args.pywb_port,
            "enable_auto_colls": False,
        }
        wsgi_app = pywb.framework.wsgi_wrappers.init_app(
            pywb.webapp.pywb_init.create_wb_router, config=conf, load_yaml=False
        )

        # disable is_hop_by_hop restrictions
        wsgiref.handlers.is_hop_by_hop = lambda x: False
        return wsgiref.simple_server.make_server(
            args.pywb_address, args.pywb_port, wsgi_app, ThreadingWSGIServer
        )

    def start(self):
        self.logger.info("starting warcprox")
        self.warcprox_controller.start()

        # XXX wait til fully started?
        self.logger.info("starting brozzler-worker")
        self.brozzler_worker.start()

        self.logger.info(
            "starting pywb", address="%s:%s" % self.pywb_httpd.server_address
        )
        threading.Thread(target=self.pywb_httpd.serve_forever).start()

        self.logger.info(
            "starting brozzler-dashboard",
            address="%s:%s" % self.dashboard_httpd.server_address,
        )
        threading.Thread(target=self.dashboard_httpd.serve_forever).start()

    def shutdown(self):
        self.logger.info("shutting down brozzler-dashboard")
        self.dashboard_httpd.shutdown()

        self.logger.info("shutting down brozzler-worker")
        self.brozzler_worker.shutdown_now()
        # brozzler-worker is fully shut down at this point

        self.logger.info("shutting down pywb")
        self.pywb_httpd.shutdown()

        self.logger.info("shutting down warcprox")
        self.warcprox_controller.shutdown()

    def wait_for_shutdown_request(self):
        try:
            while not self.stop.is_set():
                time.sleep(0.5)
        finally:
            self.shutdown()

    def _warcprox_opts(self, args):
        """
        Takes args as produced by the argument parser built by
        _build_arg_parser and builds warcprox arguments object suitable to pass
        to warcprox.main.init_controller. Copies some arguments, renames some,
        populates some with defaults appropriate for brozzler-easy, etc.
        """
        warcprox_opts = warcprox.Options()
        warcprox_opts.address = "localhost"
        # let the OS choose an available port; discover it later using
        # sock.getsockname()[1]
        warcprox_opts.port = 0
        warcprox_opts.cacert = args.cacert
        warcprox_opts.certs_dir = args.certs_dir
        warcprox_opts.directory = args.warcs_dir
        warcprox_opts.gzip = True
        warcprox_opts.prefix = "brozzler"
        warcprox_opts.size = 1000 * 1000 * 1000
        warcprox_opts.rollover_idle_time = 3 * 60
        warcprox_opts.digest_algorithm = "sha1"
        warcprox_opts.base32 = True
        warcprox_opts.stats_db_file = None
        warcprox_opts.playback_port = None
        warcprox_opts.playback_index_db_file = None
        warcprox_opts.rethinkdb_big_table_url = "rethinkdb://%s/%s/captures" % (
            args.rethinkdb_servers,
            args.rethinkdb_db,
        )
        warcprox_opts.queue_size = 500
        warcprox_opts.max_threads = None
        warcprox_opts.profile = False
        warcprox_opts.onion_tor_socks_proxy = args.onion_tor_socks_proxy
        return warcprox_opts

    def dump_state(self, signum=None, frame=None):
        state_strs = []
        for th in threading.enumerate():
            state_strs.append(str(th))
            stack = traceback.format_stack(sys._current_frames()[th.ident])
            state_strs.append("".join(stack))
        logger.warning(
            "dumping state (caught signal)", signal=signum, state="\n".join(state_strs)
        )


def main(argv=None):
    argv = argv or sys.argv
    arg_parser = _build_arg_parser(argv)
    args = arg_parser.parse_args(args=argv[1:])
    brozzler.cli.configure_logging(args)
    brozzler.chrome.check_version(args.chrome_exe)

    controller = BrozzlerEasyController(args)
    signal.signal(signal.SIGTERM, lambda a, b: controller.stop.set())
    signal.signal(signal.SIGINT, lambda a, b: controller.stop.set())
    signal.signal(signal.SIGQUIT, controller.dump_state)
    controller.start()
    controller.wait_for_shutdown_request()



"""
brozzler/__init__.py - __init__.py for brozzler package, contains some common
code

Copyright (C) 2014-2017 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import importlib.util
import logging
import threading
from importlib.metadata import version as _version

import structlog
import urlcanon

__version__ = _version("brozzler")


class ShutdownRequested(Exception):
    pass


class NothingToClaim(Exception):
    pass


class CrawlStopped(Exception):
    pass


class PageInterstitialShown(Exception):
    pass


class VideoExtractorError(Exception):
    pass


class ProxyError(Exception):
    pass


class PageConnectionError(Exception):
    pass


class ReachedTimeLimit(Exception):
    pass


class ReachedLimit(Exception):
    def __init__(self, http_error=None, warcprox_meta=None, http_payload=None):
        import json

        if http_error:
            if "warcprox-meta" in http_error.headers:
                self.warcprox_meta = json.loads(http_error.headers["warcprox-meta"])
            else:
                self.warcprox_meta = None
            self.http_payload = http_error.read()
        elif warcprox_meta:
            self.warcprox_meta = warcprox_meta
            self.http_payload = http_payload

    def __repr__(self):
        return "ReachedLimit(warcprox_meta=%r,http_payload=%r)" % (
            self.warcprox_meta if hasattr(self, "warcprox_meta") else None,
            self.http_payload if hasattr(self, "http_payload") else None,
        )

    def __str__(self):
        return self.__repr__()


# see https://github.com/internetarchive/brozzler/issues/91
def _logging_handler_handle(self, record):
    rv = self.filter(record)
    if rv:
        try:
            self.acquire()
            self.emit(record)
        finally:
            try:
                self.release()
            except:  # noqa: E722
                pass
    return rv


logging.Handler.handle = _logging_handler_handle

_behaviors = None


def behaviors(behaviors_dir=None):
    """Return list of JS behaviors loaded from YAML file.

    :param behaviors_dir: Directory containing `behaviors.yaml` and
    `js-templates/`. Defaults to brozzler dir.
    """
    import os

    import yaml

    global _behaviors
    if _behaviors is None:
        d = behaviors_dir or os.path.dirname(__file__)
        behaviors_yaml = os.path.join(d, "behaviors.yaml")
        with open(behaviors_yaml) as fin:
            _behaviors = yaml.safe_load(fin)
    return _behaviors


def behavior_script(url, template_parameters=None, behaviors_dir=None):
    """
    Returns the javascript behavior string populated with template_parameters.
    """
    import re

    logger = structlog.get_logger(logger_name=__name__)

    for behavior in behaviors(behaviors_dir=behaviors_dir):
        if re.match(behavior["url_regex"], url):
            parameters = dict()
            if "default_parameters" in behavior:
                parameters.update(behavior["default_parameters"])
            if template_parameters:
                parameters.update(template_parameters)
            template = jinja2_environment(behaviors_dir).get_template(
                behavior["behavior_js_template"]
            )
            script = template.render(parameters)
            logger.info(
                "rendering template",
                template=behavior["behavior_js_template"],
                parameters=parameters,
                url=url,
            )
            return script
    return None


class ThreadExceptionGate:
    logger = structlog.get_logger(logger_name=__module__ + "." + __qualname__)

    def __init__(self, thread):
        self.thread = thread
        self.ok_to_raise = threading.Event()
        self.pending_exception = None
        self.lock = threading.RLock()

    def __enter__(self):
        assert self.thread == threading.current_thread()
        if self.pending_exception:
            self.logger.info(
                "raising pending exception", pending_exception=self.pending_exception
            )
            tmp = self.pending_exception
            self.pending_exception = None
            raise tmp
        else:
            self.ok_to_raise.set()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        assert self.thread == threading.current_thread()
        self.ok_to_raise.clear()
        return False  # don't swallow exception

    def queue_exception(self, e):
        with self.lock:
            if self.pending_exception:
                self.logger.warning(
                    "exception already pending for thread, discarding",
                    pending_exception=self.pending_exception,
                    thread=self.thread,
                    discarded_exception=e,
                )
            else:
                self.pending_exception = e

    def __repr__(self):
        return "<ThreadExceptionGate(%s)>" % self.thread


_thread_exception_gates = {}
_thread_exception_gates_lock = threading.Lock()


def thread_exception_gate(thread=None):
    """
    Returns a `ThreadExceptionGate` for `thread` (current thread by default).

    `ThreadExceptionGate` is a context manager which allows exceptions to be
    raised from threads other than the current one, by way of `thread_raise`.

    Example:

        try:
            with thread_exception_gate():
                # do something
        except:
            # handle exception....

    If `thread_raise` is called on a thread that is not currently inside the
    `ThreadExceptionGate` context (pep340 "runtime environment"), the exception
    is queued, and raised immediately if and when the thread enters the
    context. Only one exception will be queued this way at a time, others are
    discarded.
    """
    if not thread:
        thread = threading.current_thread()

    with _thread_exception_gates_lock:
        if thread not in _thread_exception_gates:
            _thread_exception_gates[thread] = ThreadExceptionGate(thread)

    return _thread_exception_gates[thread]


thread_accept_exceptions = thread_exception_gate


def thread_raise(thread, exctype):
    """
    Raises or queues the exception `exctype` for the thread `thread`.

    See the documentation on the function `thread_exception_gate()` for more
    information.

    Adapted from http://tomerfiliba.com/recipes/Thread2/ which explains:
    "The exception will be raised only when executing python bytecode. If your
    thread calls a native/built-in blocking function, the exception will be
    raised only when execution returns to the python code."

    Raises:
        TypeError if `exctype` is not a class
        ValueError, SystemError in case of unexpected problems
    """
    import ctypes
    import inspect

    import structlog

    logger = structlog.get_logger(exctype=exctype, thread=thread)

    if not inspect.isclass(exctype):
        raise TypeError(
            "cannot raise %s, only exception types can be raised (not "
            "instances)" % exctype
        )

    gate = thread_exception_gate(thread)
    with gate.lock:
        if gate.ok_to_raise.is_set() and thread.is_alive():
            gate.ok_to_raise.clear()
            logger.info("raising exception in thread")
            res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
                ctypes.c_long(thread.ident), ctypes.py_object(exctype)
            )
            if res == 0:
                raise ValueError("invalid thread id? thread.ident=%s" % thread.ident)
            elif res != 1:
                # if it returns a number greater than one, you're in trouble,
                # and you should call it again with exc=NULL to revert the effect
                ctypes.pythonapi.PyThreadState_SetAsyncExc(thread.ident, 0)
                raise SystemError("PyThreadState_SetAsyncExc failed")
        else:
            logger.info("queueing exception for thread")
            gate.queue_exception(exctype)


def sleep(duration):
    """
    Sleeps for duration seconds in increments of 0.5 seconds.

    Use this so that the sleep can be interrupted by thread_raise().
    """
    import time

    start = time.time()
    while True:
        elapsed = time.time() - start
        if elapsed >= duration:
            break
        time.sleep(min(duration - elapsed, 0.5))


_jinja2_env = None


def jinja2_environment(behaviors_dir=None):
    global _jinja2_env
    if not _jinja2_env:
        import json
        import os

        import jinja2

        if behaviors_dir:
            _loader = jinja2.FileSystemLoader(
                os.path.join(behaviors_dir, "js-templates")
            )
        else:
            _loader = jinja2.PackageLoader("brozzler", "js-templates")
        _jinja2_env = jinja2.Environment(loader=_loader, auto_reload=False)
        _jinja2_env.filters["json"] = json.dumps
    return _jinja2_env


def _remove_query(url):
    url.question_mark = b""
    url.query = b""


# XXX chop off path after last slash??
site_surt_canon = urlcanon.Canonicalizer(urlcanon.semantic.steps + [_remove_query])


def _mdfind(identifier):
    import subprocess

    try:
        result = subprocess.check_output(
            ["mdfind", f"kMDItemCFBundleIdentifier == {identifier}"], text=True
        )
    # Just treat any errors as "couldn't find app"
    except subprocess.CalledProcessError:
        return None

    if result:
        return result.rstrip("\n")


def _suggest_default_chrome_exe_mac():
    import os

    path = None
    # Try Chromium first, then Chrome
    result = _mdfind("org.chromium.Chromium")
    if result is not None:
        path = f"{result}/Contents/MacOS/Chromium"

    result = _mdfind("com.google.Chrome")
    if result is not None:
        path = f"{result}/Contents/MacOS/Google Chrome"

    if path is not None and os.path.exists(path):
        return path

    # Fall back to default paths if mdfind couldn't find it
    # (mdfind might fail to find them even in their default paths
    # if the system has Spotlight disabled.)
    for path in [
        "/Applications/Chromium.app/Contents/MacOS/Chromium",
        "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome",
    ]:
        if os.path.exists(path):
            return path


def suggest_default_chrome_exe():
    import shutil
    import sys

    # First ask mdfind, which lets us find it in non-default paths
    if sys.platform == "darwin":
        path = _suggest_default_chrome_exe_mac()
        if path is not None:
            return path

    # "chromium-browser" is the executable on ubuntu trusty
    # https://github.com/internetarchive/brozzler/pull/6/files uses "chromium"
    # google chrome executable names taken from these packages:
    # http://www.ubuntuupdates.org/ppa/google_chrome
    for exe in [
        "chromium-browser",
        "chromium",
        "google-chrome",
        "google-chrome-stable",
        "google-chrome-beta",
        "google-chrome-unstable",
    ]:
        if shutil.which(exe):
            return exe
    return "chromium-browser"


EPOCH_UTC = datetime.datetime.fromtimestamp(0.0, tz=datetime.timezone.utc)

from brozzler.browser import Browser, BrowserPool, BrowsingException  # noqa: E402
from brozzler.robots import is_permitted_by_robots  # noqa: E402

__all__ = [
    "is_permitted_by_robots",
    "Browser",
    "BrowserPool",
    "BrowsingException",
    "sleep",
    "thread_accept_exceptions",
    "thread_raise",
    "suggest_default_chrome_exe",
]

if importlib.util.find_spec("doublethink"):
    # All of these imports use doublethink for real and are unsafe
    # to do if doublethink is unavailable.
    from brozzler.frontier import RethinkDbFrontier  # noqa: F401
    from brozzler.model import (
        InvalidJobConf,  # noqa: F401
        Job,  # noqa: F401
        Page,  # noqa: F401
        Site,  # noqa: F401
        new_job,  # noqa: F401
        new_job_file,  # noqa: F401
        new_site,  # noqa: F401
    )
    from brozzler.worker import BrozzlerWorker  # noqa: F401

    __all__.extend(
        [
            "Page",
            "BrozzlerWorker",
            "RethinkDbFrontier",
            "Site",
            "new_job",
            "new_site",
            "Job",
            "new_job_file",
            "InvalidJobConf",
        ]
    )

# we could make this configurable if there's a good reason
MAX_PAGE_FAILURES = 3



"""
brozzler/worker.py - BrozzlerWorker brozzles pages from the frontier, meaning
it runs yt-dlp on them, browses them and runs behaviors if appropriate,
scopes and adds outlinks to the frontier

Copyright (C) 2014-2025 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
import importlib.util
import io
import json
import socket
import threading
import time
import urllib.error
import urllib.request

import doublethink
import PIL.Image
import requests
import rethinkdb as rdb
import structlog
import urlcanon
import urllib3
from requests.structures import CaseInsensitiveDict
from urllib3.exceptions import ProxyError, TimeoutError

import brozzler
import brozzler.browser
from brozzler.model import VideoCaptureOptions
from brozzler.ssl import CustomSSLContextHTTPAdapter, permissive_ssl_context

from . import metrics

r = rdb.RethinkDB()


def _get_with_legacy_renegotiation(url, **kwargs) -> requests.Response:
    """
    Performs a get request with unsafe legacy renegotiation enabled.
    """
    session = requests.session()
    ctx = permissive_ssl_context()
    if not kwargs.get("verify", True):
        ctx.check_hostname = False
    session.mount("https://", CustomSSLContextHTTPAdapter(ctx))
    return session.get(url, **kwargs)


def _get_with_permissive_fallback(url, **kwargs) -> requests.Response:
    """
    Attempts to perform a request with default SSL settings, then falls
    back to a request with unsafe legacy renegotiation enabled if
    necessary.
    """
    try:
        return requests.get(url, **kwargs)
    except requests.exceptions.SSLError as e:
        if (
            e.__context__
            and e.__context__.__context__
            and e.__context__.__context__.__context__
            and e.__context__.__context__.__context__.reason
            == "UNSAFE_LEGACY_RENEGOTIATION_DISABLED"
        ):
            logger = structlog.get_logger(logger_name=__name__)
            logger.info(
                "request failed with legacy renegotiation disabled; "
                "retrying with it enabled",
                url=url,
            )
            return _get_with_legacy_renegotiation(url, **kwargs)
        raise


class BrozzlerWorker:
    logger = structlog.get_logger(logger_name=__module__ + "." + __qualname__)

    # 3⅓ min heartbeat interval => 10 min ttl
    # This is kind of a long time, because `frontier.claim_sites()`, which runs
    # in the same thread as the heartbeats, can take a while on a busy brozzler
    # cluster with slow rethinkdb.
    HEARTBEAT_INTERVAL = 200.0
    SITE_SESSION_MINUTES = 15
    HEADER_REQUEST_TIMEOUT = 60
    FETCH_URL_TIMEOUT = 60

    def __init__(
        self,
        frontier,
        service_registry=None,
        ytdlp_proxy_endpoints=None,
        max_browsers=1,
        chrome_exe="chromium-browser",
        warcprox_auto=False,
        proxy=None,
        headless=True,
        skip_extract_outlinks=False,
        skip_visit_hashtags=False,
        skip_youtube_dl=False,
        ytdlp_tmpdir="/tmp",
        simpler404=False,
        screenshot_full_page=False,
        page_timeout=300,
        behavior_timeout=300,
        extract_outlinks_timeout=60,
        download_throughput=-1,
        stealth=False,
        window_height=900,
        window_width=1400,
        metrics_port=0,
        registry_url=None,
        env=None,
        worker_id=None,
    ):
        self._frontier = frontier
        self._service_registry = service_registry
        self._ytdlp_proxy_endpoints = ytdlp_proxy_endpoints
        self._max_browsers = max_browsers

        self._warcprox_auto = warcprox_auto
        self._proxy = proxy
        assert not (warcprox_auto and proxy)
        self._proxy_is_warcprox = None
        self._headless = headless
        self._skip_extract_outlinks = skip_extract_outlinks
        self._skip_visit_hashtags = skip_visit_hashtags
        self._skip_youtube_dl = skip_youtube_dl

        if worker_id is not None:
            self.logger = self.logger.bind(worker_id=worker_id)

        # We definitely shouldn't ytdlp if the optional extra is missing
        if not importlib.util.find_spec("yt_dlp"):
            self.logger.info(
                "optional yt-dlp extra not installed; setting skip_youtube_dl to True"
            )
            self._skip_youtube_dl = True

        self._ytdlp_tmpdir = ytdlp_tmpdir
        self._simpler404 = simpler404
        self._screenshot_full_page = screenshot_full_page
        self._page_timeout = page_timeout
        self._behavior_timeout = behavior_timeout
        self._extract_outlinks_timeout = extract_outlinks_timeout
        self._download_throughput = download_throughput
        self._window_height = window_height
        self._window_width = window_width
        self._stealth = stealth
        self._metrics_port = metrics_port
        self._registry_url = registry_url
        self._env = env

        self._browser_pool = brozzler.browser.BrowserPool(
            max_browsers, chrome_exe=chrome_exe, ignore_cert_errors=True
        )
        self._browsing_threads = set()
        self._browsing_threads_lock = threading.Lock()

        self._thread = None
        self._start_stop_lock = threading.Lock()
        self._shutdown = threading.Event()

        # set up metrics
        if self._metrics_port > 0:
            metrics.register_prom_metrics(
                self._metrics_port, self._registry_url, self._env
            )
        else:
            self.logger.warning(
                "not starting prometheus scrape endpoint: metrics_port is undefined"
            )

    def _choose_warcprox(self):
        warcproxes = self._service_registry.available_services("warcprox")
        if not warcproxes:
            return None
        # .group('proxy').count() makes this query about 99% more efficient
        reql = (
            self._frontier.rr.table("sites")
            .between(
                ["ACTIVE", r.minval],
                ["ACTIVE", r.maxval],
                index="sites_last_disclaimed",
            )
            .group("proxy")
            .count()
        )
        # returns results like
        # {
        #    "wbgrp-svc030.us.archive.org:8000": 148,
        #    "wbgrp-svc030.us.archive.org:8001": 145
        # }
        proxy_scoreboard = dict(reql.run())
        for warcprox in warcproxes:
            address = "%s:%s" % (warcprox["host"], warcprox["port"])
            warcprox["assigned_sites"] = proxy_scoreboard.get(address, 0)
        warcproxes.sort(
            key=lambda warcprox: (warcprox["assigned_sites"], warcprox["load"])
        )
        # XXX make this heuristic more advanced?
        return warcproxes[0]

    def _proxy_for(self, site):
        if self._proxy:
            return self._proxy
        elif site.proxy:
            return site.proxy
        elif self._warcprox_auto:
            svc = self._choose_warcprox()
            if svc is None:
                raise brozzler.ProxyError(
                    "no available instances of warcprox in the service registry"
                )
            site.proxy = "%s:%s" % (svc["host"], svc["port"])
            site.save()
            self.logger.info(
                "chose warcprox instance from service registry",
                instance=site.proxy,
                registry=site,
            )
            return site.proxy
        return None

    def _using_warcprox(self, site):
        if self._proxy:
            if self._proxy_is_warcprox is None:
                try:
                    response = requests.get("http://%s/status" % self._proxy)
                    status = json.loads(response.text)
                    self._proxy_is_warcprox = status["role"] == "warcprox"
                except Exception:
                    self._proxy_is_warcprox = False
                self.logger.info(
                    "%s %s warcprox",
                    self._proxy,
                    "IS" if self._proxy_is_warcprox else "IS NOT",
                )
            return self._proxy_is_warcprox
        else:
            # I should have commented when I originally wrote this code, but I
            # think this works because `site.proxy` is only set when the proxy
            # is warcprox
            return bool(site.proxy or self._warcprox_auto)

    def _warcprox_write_record(
        self,
        warcprox_address,
        url,
        warc_type,
        content_type,
        payload,
        extra_headers=None,
    ):
        headers = {"Content-Type": content_type, "WARC-Type": warc_type, "Host": "N/A"}
        if extra_headers:
            headers.update(extra_headers)
        request = urllib.request.Request(
            url, method="WARCPROX_WRITE_RECORD", headers=headers, data=payload
        )

        # XXX setting request.type="http" is a hack to stop urllib from trying
        # to tunnel if url is https
        request.type = "http"
        request.set_proxy(warcprox_address, "http")

        try:
            with urllib.request.urlopen(request, timeout=600) as response:
                if response.getcode() != 204:
                    self.logger.warning(
                        "got unexpected response on warcprox "
                        "WARCPROX_WRITE_RECORD request (expected 204)",
                        code=response.getcode(),
                        reason=response.reason,
                    )
                return request, response
        except urllib.error.HTTPError as e:
            self.logger.warning(
                "got unexpected response on warcprox "
                "WARCPROX_WRITE_RECORD request (expected 204)",
                code=e.getcode(),
                reason=e.info(),
            )
            return request, None
        except urllib.error.URLError as e:
            raise brozzler.ProxyError(
                "proxy error on WARCPROX_WRITE_RECORD %s" % url
            ) from e
        except ConnectionError as e:
            raise brozzler.ProxyError(
                "proxy error on WARCPROX_WRITE_RECORD %s" % url
            ) from e

    def thumb_jpeg(self, full_jpeg):
        """Create JPEG thumbnail."""
        img = PIL.Image.open(io.BytesIO(full_jpeg))
        thumb_width = 300
        thumb_height = (thumb_width / img.size[0]) * img.size[1]
        img.thumbnail((thumb_width, thumb_height))
        out = io.BytesIO()
        img.save(out, "jpeg", quality=95)
        return out.getbuffer()

    def should_ytdlp(self, logger, site, page, page_status):
        # called only after we've passed needs_browsing() check

        if page_status != 200:
            logger.info("skipping ytdlp: non-200 page status", page_status=page_status)
            return False
        if site.video_capture in [
            VideoCaptureOptions.DISABLE_VIDEO_CAPTURE.value,
            VideoCaptureOptions.DISABLE_YTDLP_CAPTURE.value,
        ]:
            logger.info("skipping ytdlp: site has video capture disabled")
            return False

        ytdlp_url = page.redirect_url if page.redirect_url else page.url

        if "chrome-error:" in ytdlp_url:
            return False

        return True

    @metrics.brozzler_page_processing_duration_seconds.time()
    @metrics.brozzler_in_progress_pages.track_inprogress()
    def brozzle_page(
        self,
        browser,
        site,
        page,
        on_screenshot=None,
        on_request=None,
        enable_youtube_dl=True,
    ):
        page_logger = self.logger.bind(page=page)
        page_logger.info("brozzling")
        outlinks = set()

        page_headers = self._get_page_headers(site, page)

        if not self._needs_browsing(page_headers):
            page_logger.info("needs fetch")
            if site.pdfs_only and not self._is_pdf(page_headers):
                page_logger.info("skipping non-PDF content: PDFs only option enabled")
            elif site.video_capture in [
                VideoCaptureOptions.DISABLE_VIDEO_CAPTURE.value,
                VideoCaptureOptions.BLOCK_VIDEO_MIME_TYPES.value,
            ] and self._is_media_type(page_headers):
                page_logger.info(
                    "skipping audio/video content: video MIME type capture disabled for site"
                )
            else:
                self._fetch_url(site, page=page)
        else:
            page_logger.info("needs browsing")
            try:
                browser_outlinks = self._browse_page(
                    browser, site, page, on_screenshot, on_request
                )
                outlinks.update(browser_outlinks)
                status_code = browser.websock_thread.page_status
                if status_code in [502, 504] or (
                    page.redirect_url and page.redirect_url.startswith("chrome-error:")
                ):
                    self.logger.warning("Chrome Error page encountered", page=page)
                    raise brozzler.PageConnectionError()
            except brozzler.PageInterstitialShown:
                page_logger.info("page interstitial shown (http auth)")
                status_code = -1

            if enable_youtube_dl and self.should_ytdlp(
                page_logger, site, page, status_code
            ):
                try:
                    from . import ydl

                    ydl_outlinks = ydl.do_youtube_dl(
                        self, site, page, self._ytdlp_proxy_endpoints
                    )
                    metrics.brozzler_ydl_urls_checked.inc(1)
                    outlinks.update(ydl_outlinks)
                except brozzler.ReachedLimit:
                    raise
                except brozzler.ShutdownRequested:
                    raise
                except brozzler.ProxyError:
                    raise
                except brozzler.VideoExtractorError:
                    self.logger.exception("error extracting video info")
                except Exception as e:
                    if (
                        hasattr(e, "exc_info")
                        and len(e.exc_info) >= 2
                        and hasattr(e.exc_info[1], "code")
                        and e.exc_info[1].code == 430
                    ):
                        self.logger.info(
                            "youtube-dl encountered an error",
                            code=e.exc_info[1].code,
                            message=e.exc_info[1].msg,
                            url=page.url,
                        )
                    else:
                        self.logger.exception("youtube_dl raised exception", page=page)
        return outlinks

    @metrics.brozzler_header_processing_duration_seconds.time()
    @metrics.brozzler_in_progress_headers.track_inprogress()
    def _get_page_headers(self, site, page):
        url_logger = self.logger.bind(url=page.url)
        # bypassing warcprox, requests' stream=True defers downloading the body of the response
        # see https://docs.python-requests.org/en/latest/user/advanced/#body-content-workflow
        try:
            user_agent = site.get("user_agent")
            headers = {"User-Agent": user_agent} if user_agent else {}
            url_logger.info("getting page headers")
            with _get_with_permissive_fallback(
                page.url,
                stream=True,
                verify=False,
                headers=headers,
                timeout=self.HEADER_REQUEST_TIMEOUT,
            ) as r:
                return r.headers
        except requests.exceptions.Timeout:
            url_logger.warning("Timed out trying to get headers", exc_info=True)
        except requests.exceptions.RequestException:
            url_logger.warning("Failed to get headers", exc_info=True)
        return {}

    def _needs_browsing(self, page_headers) -> bool:
        return not (
            "content-type" in page_headers
            and "html" not in page_headers["content-type"]
        )

    def _is_media_type(self, page_headers) -> bool:
        """
        Determines if the page's Content-Type header specifies that it contains
        audio or video.
        """
        return "content-type" in page_headers and (
            "video" in page_headers["content-type"]
            or "audio" in page_headers["content-type"]
            # https://github.com/gsuberland/UMP_Format/blob/main/UMP_Format.md
            or page_headers["content-type"] == "application/vnd.yt-ump"
        )

    def _is_pdf(self, page_headers) -> bool:
        """
        Determines if the page's Content-Type header specifies that it is a PDF.
        """
        return (
            "content-type" in page_headers
            and "application/pdf" in page_headers["content-type"]
        )

    @metrics.brozzler_browsing_duration_seconds.time()
    @metrics.brozzler_in_progress_browses.track_inprogress()
    def _browse_page(self, browser, site, page, on_screenshot=None, on_request=None):
        def update_page_metrics(page, outlinks):
            """Update page-level Prometheus metrics."""
            metrics.brozzler_last_page_crawled_time.set_to_current_time()
            metrics.brozzler_pages_crawled.inc(1)
            metrics.brozzler_outlinks_found.inc(len(outlinks))

        def _on_screenshot(screenshot_jpeg):
            if on_screenshot:
                on_screenshot(screenshot_jpeg)
            if self._using_warcprox(site):
                self.logger.info(
                    "sending WARCPROX_WRITE_RECORD request",
                    proxy=self._proxy_for(site),
                    screenshot_for_page=page,
                )
                thumbnail_jpeg = self.thumb_jpeg(screenshot_jpeg)
                self._warcprox_write_record(
                    warcprox_address=self._proxy_for(site),
                    url="screenshot:%s" % str(urlcanon.semantic(page.url)),
                    warc_type="resource",
                    content_type="image/jpeg",
                    payload=screenshot_jpeg,
                    extra_headers=site.extra_headers(page),
                )
                self._warcprox_write_record(
                    warcprox_address=self._proxy_for(site),
                    url="thumbnail:%s" % str(urlcanon.semantic(page.url)),
                    warc_type="resource",
                    content_type="image/jpeg",
                    payload=thumbnail_jpeg,
                    extra_headers=site.extra_headers(page),
                )

        def _on_response(chrome_msg):
            if (
                "params" in chrome_msg
                and "response" in chrome_msg["params"]
                and "mimeType" in chrome_msg["params"]["response"]
                and chrome_msg["params"]["response"]
                .get("mimeType", "")
                .startswith("video/")
                # skip manifests of DASH segmented video -
                # see https://github.com/internetarchive/brozzler/pull/70
                and chrome_msg["params"]["response"]["mimeType"]
                != "video/vnd.mpeg.dash.mpd"
                and chrome_msg["params"]["response"].get("status") in (200, 206)
            ):
                video = {
                    "blame": "browser",
                    "url": chrome_msg["params"]["response"].get("url"),
                    "response_code": chrome_msg["params"]["response"]["status"],
                    "content-type": chrome_msg["params"]["response"]["mimeType"],
                }
                response_headers = CaseInsensitiveDict(
                    chrome_msg["params"]["response"]["headers"]
                )
                if "content-length" in response_headers:
                    video["content-length"] = int(response_headers["content-length"])
                if "content-range" in response_headers:
                    video["content-range"] = response_headers["content-range"]
                self.logger.debug("embedded video", video=video)
                if "videos" not in page:
                    page.videos = []
                page.videos.append(video)

        sw_fetched = set()

        def _on_service_worker_version_updated(chrome_msg):
            # https://github.com/internetarchive/brozzler/issues/140
            self.logger.debug("service worker updated", chrome_msg=chrome_msg)
            if chrome_msg.get("params", {}).get("versions"):
                url = chrome_msg.get("params", {}).get("versions")[0].get("scriptURL")
                if url and url.startswith("http") and url not in sw_fetched:
                    self.logger.info("fetching service worker script", url=url)
                    self._fetch_url(site, url=url)
                    sw_fetched.add(url)

        if not browser.is_running():
            browser.start(
                proxy=self._proxy_for(site),
                cookie_db=site.get("cookie_db"),
                window_height=self._window_height,
                window_width=self._window_width,
                headless=self._headless,
            )
        page.clear_redirect()
        final_page_url, outlinks = browser.browse_page(
            page.url,
            extra_headers=site.extra_headers(page),
            behavior_parameters=site.get("behavior_parameters"),
            username=site.get("username"),
            password=site.get("password"),
            user_agent=site.get("user_agent"),
            on_screenshot=_on_screenshot,
            on_response=_on_response,
            on_request=on_request,
            on_service_worker_version_updated=_on_service_worker_version_updated,
            hashtags=page.hashtags,
            skip_extract_outlinks=self._skip_extract_outlinks,
            skip_visit_hashtags=self._skip_visit_hashtags,
            skip_youtube_dl=self._skip_youtube_dl,
            ytdlp_tmpdir=self._ytdlp_tmpdir,
            simpler404=self._simpler404,
            screenshot_full_page=self._screenshot_full_page,
            page_timeout=self._page_timeout,
            behavior_timeout=self._behavior_timeout,
            extract_outlinks_timeout=self._extract_outlinks_timeout,
            download_throughput=self._download_throughput,
            stealth=self._stealth,
        )
        if final_page_url != page.url:
            page.note_redirect(final_page_url)
        update_page_metrics(page, outlinks)
        return outlinks

    def _fetch_url(self, site, url=None, page=None):
        proxy_url = self._proxy_for(site)

        if page:
            url = page.url

        if proxy_url:
            http = urllib3.ProxyManager("http://%s" % proxy_url, cert_reqs="CERT_NONE")
        else:
            http = urllib3.PoolManager()

        user_agent = site.get("user_agent")
        headers = {"User-Agent": user_agent} if user_agent else {}
        headers.update(site.extra_headers(page))

        self.logger.info("fetching url", url=url)
        try:
            # response is ignored
            http.request(
                "GET",
                url,
                headers=headers,
                timeout=self.FETCH_URL_TIMEOUT,
                retries=False,
            )
            self.logger.info("Completed fetching url", url=url)
        except TimeoutError as e:
            self.logger.warning("Timed out fetching url", url=url)
            raise brozzler.PageConnectionError() from e
        except ProxyError as e:
            raise brozzler.ProxyError("proxy error fetching %s" % url) from e
        except urllib3.exceptions.RequestError as e:
            self.logger.warning("Failed to fetch url", url=url, exc_info=True)
            raise brozzler.PageConnectionError() from e

    def brozzle_site(self, browser, site):
        site_logger = self.logger.bind(site=site)
        try:
            site.last_claimed_by = "%s:%s" % (socket.gethostname(), browser.chrome.port)
            site.save()
            start = time.time()
            page = None
            self._frontier.enforce_time_limit(site)
            self._frontier.honor_stop_request(site)
            # _proxy_for() call in log statement can raise brozzler.ProxyError
            # which is why we honor time limit and stop request first☝🏻
            site_logger.info("brozzling site", proxy=self._proxy_for(site))
            while time.time() - start < self.SITE_SESSION_MINUTES * 60:
                site.refresh()
                self._frontier.enforce_time_limit(site)
                self._frontier.honor_stop_request(site)
                page = self._frontier.claim_page(
                    site, "%s:%s" % (socket.gethostname(), browser.chrome.port)
                )

                if page.needs_robots_check and not brozzler.is_permitted_by_robots(
                    site, page.url, self._proxy_for(site)
                ):
                    self.logger.warning("page is blocked by robots.txt", url=page.url)
                    page.blocked_by_robots = True
                    self._frontier.completed_page(site, page)
                else:
                    outlinks = self.brozzle_page(
                        browser, site, page, enable_youtube_dl=not self._skip_youtube_dl
                    )
                    self._frontier.completed_page(site, page)
                    self._frontier.scope_and_schedule_outlinks(site, page, outlinks)
                    if browser.is_running():
                        site.cookie_db = browser.chrome.persist_and_read_cookie_db()

                page = None
        except brozzler.ShutdownRequested:
            self.logger.info("shutdown requested")
        except brozzler.NothingToClaim:
            site_logger.info("no pages left for site")
        except brozzler.ReachedLimit as e:
            self._frontier.reached_limit(site, e)
        except brozzler.ReachedTimeLimit:
            self._frontier.finished(site, "FINISHED_TIME_LIMIT")
        except brozzler.CrawlStopped:
            self._frontier.finished(site, "FINISHED_STOP_REQUESTED")
        # except brozzler.browser.BrowsingAborted:
        #     self.logger.info("{} shut down".format(browser))
        except brozzler.ProxyError:
            if self._warcprox_auto:
                self.logger.exception(
                    "proxy error, will try to choose a "
                    "healthy instance next time site is brozzled",
                    site_proxy=site.proxy,
                )
                site.proxy = None
            else:
                # using brozzler-worker --proxy, nothing to do but try the
                # same proxy again next time
                self.logger.exception("proxy error", self_proxy=self._proxy)
        except (brozzler.PageConnectionError, Exception) as e:
            if isinstance(e, brozzler.PageConnectionError):
                site_logger.exception(
                    "Page status code possibly indicates connection failure between host and warcprox",
                    page=page,
                )
            else:
                site_logger.exception("unexpected exception", page=page)
            if page:
                # Calculate backoff in seconds based on number of failed attempts.
                # Minimum of 60, max of 135 giving delays of 60, 90, 135, 135...
                retry_delay = min(135, 60 * (1.5 ** (page.failed_attempts or 0)))
                page.retry_after = doublethink.utcnow() + datetime.timedelta(
                    seconds=retry_delay
                )
                page.failed_attempts = (page.failed_attempts or 0) + 1
                if page.failed_attempts >= brozzler.MAX_PAGE_FAILURES:
                    self.logger.info(
                        'marking page "completed" after several unexpected '
                        "exceptions attempting to brozzle",
                        failed_attempts=page.failed_attempts,
                        page=page,
                    )
                    self._frontier.completed_page(site, page)
                    page = None
                else:
                    page.save()
        finally:
            if start:
                site.active_brozzling_time = (
                    (site.active_brozzling_time or 0) + time.time() - start
                )
            self._frontier.disclaim_site(site, page)

    def _brozzle_site_thread_target(self, browser, site):
        try:
            self.brozzle_site(browser, site)
        finally:
            browser.stop()
            self._browser_pool.release(browser)
            with self._browsing_threads_lock:
                self._browsing_threads.remove(threading.current_thread())

    def _service_heartbeat(self):
        if hasattr(self, "status_info"):
            status_info = self.status_info
        else:
            status_info = {
                "role": "brozzler-worker",
                "ttl": self.HEARTBEAT_INTERVAL * 3,
            }
        status_info["load"] = (
            1.0 * self._browser_pool.num_in_use() / self._browser_pool.size
        )
        status_info["browser_pool_size"] = self._browser_pool.size
        status_info["browsers_in_use"] = self._browser_pool.num_in_use()

        try:
            self.status_info = self._service_registry.heartbeat(status_info)
            self.logger.debug("status in service registry", status=self.status_info)
        except r.ReqlError:
            self.logger.exception(
                "failed to send heartbeat and update service registry",
                info=status_info,
            )

    def _service_heartbeat_if_due(self):
        """Sends service registry heartbeat if due"""
        due = False
        if self._service_registry:
            if not hasattr(self, "status_info"):
                due = True
            else:
                d = doublethink.utcnow() - self.status_info["last_heartbeat"]
                due = d.total_seconds() > self.HEARTBEAT_INTERVAL

        if due:
            self._service_heartbeat()

    def _start_browsing_some_sites(self):
        """
        Starts browsing some sites.

        Raises:
            NoBrowsersAvailable if none available
        """
        # acquire_multi() raises NoBrowsersAvailable if none available
        browsers = self._browser_pool.acquire_multi(
            (self._browser_pool.num_available() + 1) // 2
        )
        try:
            sites = self._frontier.claim_sites(len(browsers))
        except:
            self._browser_pool.release_all(browsers)
            raise

        for i in range(len(browsers)):
            if i < len(sites):
                th = threading.Thread(
                    target=self._brozzle_site_thread_target,
                    args=(browsers[i], sites[i]),
                    name="BrozzlingThread:%s" % browsers[i].chrome.port,
                    daemon=True,
                )
                with self._browsing_threads_lock:
                    self._browsing_threads.add(th)
                th.start()
            else:
                self._browser_pool.release(browsers[i])

    def run(self):
        self.logger.warn("brozzler %s - brozzler-worker starting", brozzler.__version__)
        last_nothing_to_claim = 0
        try:
            while not self._shutdown.is_set():
                self._service_heartbeat_if_due()
                if time.time() - last_nothing_to_claim > 20:
                    try:
                        self._start_browsing_some_sites()
                    except brozzler.browser.NoBrowsersAvailable:
                        self.logger.debug(
                            "all browsers are in use", max_browsers=self._max_browsers
                        )
                    except brozzler.NothingToClaim:
                        last_nothing_to_claim = time.time()
                        self.logger.debug(
                            "nothing to claim, all available active sites "
                            "are already claimed by a brozzler worker"
                        )
                time.sleep(0.5)

            self.logger.warn("shutdown requested")
        except r.ReqlError:
            self.logger.exception("caught rethinkdb exception, will try to proceed")
        except brozzler.ShutdownRequested:
            self.logger.info("shutdown requested")
        except:  # noqa: E722
            self.logger.critical(
                "thread exiting due to unexpected exception", exc_info=True
            )
        finally:
            if self._service_registry and hasattr(self, "status_info"):
                try:
                    self._service_registry.unregister(self.status_info["id"])
                except:  # noqa: E722
                    self.logger.exception("failed to unregister from service registry")

            self.logger.info(
                "shutting down brozzling threads",
                thread_count=len(self._browsing_threads),
            )
            with self._browsing_threads_lock:
                for th in self._browsing_threads:
                    if th.is_alive():
                        brozzler.thread_raise(th, brozzler.ShutdownRequested)
            self._browser_pool.shutdown_now()
            # copy to avoid "RuntimeError: Set changed size during iteration"
            thredz = set(self._browsing_threads)
            for th in thredz:
                th.join()

    def start(self):
        with self._start_stop_lock:
            if self._thread:
                self.logger.warning(
                    "ignoring start request because self._thread is not None"
                )
                return
            self._thread = threading.Thread(target=self.run, name="BrozzlerWorker")
            self.logger = self.logger.bind(thread=self._thread)
            self._thread.start()

    def shutdown_now(self):
        self.stop()

    def stop(self):
        self._shutdown.set()

    def is_alive(self):
        return self._thread and self._thread.is_alive()



#!/usr/bin/env python
"""
test_brozzling.py - XXX explain

Copyright (C) 2016-2018 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import argparse
import http.server
import json
import logging
import os
import socket
import threading
import time
import urllib

import pytest

import brozzler.cli

arg_parser = argparse.ArgumentParser()
brozzler.cli.add_common_options(arg_parser)
args = arg_parser.parse_args([])
args.log_level = logging.INFO
brozzler.cli.configure_logging(args)

WARCPROX_META_420 = {
    "stats": {
        "test_limits_bucket": {
            "total": {"urls": 0, "wire_bytes": 0},
            "new": {"urls": 0, "wire_bytes": 0},
            "revisit": {"urls": 0, "wire_bytes": 0},
            "bucket": "test_limits_bucket",
        }
    },
    "reached-limit": {"test_limits_bucket/total/urls": 0},
}


@pytest.fixture(scope="module")
def httpd(request):
    class RequestHandler(http.server.SimpleHTTPRequestHandler):
        def __init__(self, *args, **kwargs):
            self.extensions_map[".mpd"] = "video/vnd.mpeg.dash.mpd"
            http.server.SimpleHTTPRequestHandler.__init__(self, *args, **kwargs)

        def do_GET(self):
            if self.path == "/420":
                self.send_response(420, "Reached limit")
                self.send_header("Connection", "close")
                self.send_header("Warcprox-Meta", json.dumps(WARCPROX_META_420))
                payload = b"request rejected by warcprox: reached limit test_limits_bucket/total/urls=0\n"
                self.send_header("Content-Type", "text/plain;charset=utf-8")
                self.send_header("Content-Length", len(payload))
                self.end_headers()
                self.wfile.write(payload)
            elif self.path == "/401":
                self.send_response(401)
                self.send_header("WWW-Authenticate", 'Basic realm="Test"')
                self.send_header("Content-type", "text/html")
                self.end_headers()
                self.wfile.write(self.headers.get("Authorization", b""))
                self.wfile.write(b"not authenticated")
            elif self.path == "/sleep-3s":
                self.send_response(200)
                self.end_headers()
                time.sleep(3)
                self.wfile.write(b"ok")
            else:
                super().do_GET()

        def do_POST(self):
            if self.path == "/login-action":
                self.send_response(200)
                payload = b"login successful\n"
                self.send_header("Content-Type", "text/plain;charset=utf-8")
                self.send_header("Content-Length", len(payload))
                self.end_headers()
                self.wfile.write(payload)
            else:
                super().do_POST()

    # SimpleHTTPRequestHandler always uses CWD so we have to chdir
    os.chdir(os.path.join(os.path.dirname(__file__), "htdocs"))

    httpd = http.server.HTTPServer(("localhost", 0), RequestHandler)
    httpd_thread = threading.Thread(name="httpd", target=httpd.serve_forever)
    httpd_thread.start()

    def fin():
        httpd.shutdown()
        httpd.server_close()
        httpd_thread.join()

    request.addfinalizer(fin)

    return httpd


def test_httpd(httpd):
    """
    Tests that our http server is working as expected, and that two fetches
    of the same url return the same payload, proving it can be used to test
    deduplication.
    """
    payload1 = None
    url = "http://localhost:%s/site1/file1.txt" % httpd.server_port
    with urllib.request.urlopen(url) as response:
        assert response.status == 200
        payload1 = response.read()
        assert payload1

    with urllib.request.urlopen(url) as response:
        assert response.status == 200
        payload2 = response.read()
        assert payload2

    assert payload1 == payload2

    url = "http://localhost:%s/420" % httpd.server_port
    with pytest.raises(urllib.error.HTTPError) as excinfo:
        urllib.request.urlopen(url)
    assert excinfo.value.getcode() == 420


def test_aw_snap_hes_dead_jim():
    chrome_exe = brozzler.suggest_default_chrome_exe()
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        with pytest.raises(brozzler.BrowsingException):
            browser.browse_page("chrome://crash")


# chromium's 401 handling changed???
@pytest.mark.xfail
def test_page_interstitial_exception(httpd):
    chrome_exe = brozzler.suggest_default_chrome_exe()
    url = "http://localhost:%s/401" % httpd.server_port
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        with pytest.raises(brozzler.PageInterstitialShown):
            browser.browse_page(url)


def test_on_response(httpd):
    response_urls = []

    def on_response(msg):
        response_urls.append(msg["params"]["response"]["url"])

    chrome_exe = brozzler.suggest_default_chrome_exe()
    url = "http://localhost:%s/site3/page.html" % httpd.server_port
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        browser.browse_page(url, on_response=on_response)
    assert response_urls[0] == "http://localhost:%s/site3/page.html" % httpd.server_port
    assert (
        response_urls[1] == "http://localhost:%s/site3/brozzler.svg" % httpd.server_port
    )
    assert response_urls[2] == "http://localhost:%s/favicon.ico" % httpd.server_port


def test_420(httpd):
    chrome_exe = brozzler.suggest_default_chrome_exe()
    url = "http://localhost:%s/420" % httpd.server_port
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        with pytest.raises(brozzler.ReachedLimit) as excinfo:
            browser.browse_page(url)
        assert excinfo.value.warcprox_meta == WARCPROX_META_420


def test_js_dialogs(httpd):
    chrome_exe = brozzler.suggest_default_chrome_exe()
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        # before commit d2ed6b97a24 these would hang and eventually raise
        # brozzler.browser.BrowsingTimeout, which would cause this test to fail
        browser.browse_page("http://localhost:%s/site4/alert.html" % httpd.server_port)
        browser.browse_page(
            "http://localhost:%s/site4/confirm.html" % httpd.server_port
        )
        browser.browse_page("http://localhost:%s/site4/prompt.html" % httpd.server_port)
        # XXX print dialog unresolved
        # browser.browse_page(
        #         'http://localhost:%s/site4/print.html' % httpd.server_port)


def test_page_videos(httpd):
    # test depends on behavior of youtube-dl and chromium, could fail and need
    # to be adjusted on youtube-dl or chromium updates
    chrome_exe = brozzler.suggest_default_chrome_exe()
    worker = brozzler.BrozzlerWorker(None)
    site = brozzler.Site(None, {})
    page = brozzler.Page(
        None, {"url": "http://localhost:%s/site6/" % httpd.server_port}
    )
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        worker.brozzle_page(browser, site, page)
    assert page.videos
    assert len(page.videos) == 1
    # assert page.videos[0] == {
    #     "blame": "youtube-dl",
    #     "response_code": 200,
    #     "content-length": 383631,
    #     "content-type": "video/mp4",
    #     "url": "http://localhost:%s/site6/small.mp4" % httpd.server_port,
    # }
    # assert page.videos[1] == {
    #     "blame": "youtube-dl",
    #     "content-length": 92728,
    #     "content-type": "video/webm",
    #     "response_code": 200,
    #     "url": "http://localhost:%s/site6/small-video_280x160_100k.webm"
    #     % httpd.server_port,
    # }
    # assert page.videos[2] == {
    #     "blame": "youtube-dl",
    #     "content-length": 101114,
    #     "content-type": "video/webm",
    #     "response_code": 200,
    #     "url": "http://localhost:%s/site6/small-audio.webm" % httpd.server_port,
    # }
    assert page.videos[0] == {
        "blame": "browser",
        # 'response_code': 206,
        # 'content-range': 'bytes 0-229454/229455',
        "response_code": 200,
        "content-length": 229455,
        "content-type": "video/webm",
        "url": "http://localhost:%s/site6/small.webm" % httpd.server_port,
    }


def test_extract_outlinks(httpd):
    chrome_exe = brozzler.suggest_default_chrome_exe()
    worker = brozzler.BrozzlerWorker(None)
    site = brozzler.Site(None, {})
    page = brozzler.Page(
        None, {"url": "http://localhost:%s/site8/" % httpd.server_port}
    )
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        outlinks = worker.brozzle_page(browser, site, page)
    assert outlinks == {
        "http://example.com/offsite",
        "http://localhost:%s/site8/baz/zuh" % httpd.server_port,
        "http://localhost:%s/site8/fdjisapofdjisap#1" % httpd.server_port,
        "http://localhost:%s/site8/fdjisapofdjisap#2" % httpd.server_port,
    }


def test_proxy_down():
    """
    Test that browsing raises `brozzler.ProxyError` when proxy is down.

    See also `test_proxy_down` in test_units.py.

    Tests two different kinds of connection error:
    - nothing listening the port (nobody listens on on port 4 :))
    - port bound but not accepting connections
    """
    sock = socket.socket()
    sock.bind(("127.0.0.1", 0))
    for not_listening_proxy in ("127.0.0.1:4", "127.0.0.1:%s" % sock.getsockname()[1]):
        site = brozzler.Site(None, {"seed": "http://example.com/"})
        page = brozzler.Page(None, {"url": "http://example.com/"})

        worker = brozzler.BrozzlerWorker(frontier=None, proxy=not_listening_proxy)
        chrome_exe = brozzler.suggest_default_chrome_exe()

        with brozzler.Browser(chrome_exe=chrome_exe) as browser:
            browser.stop()  # We're manually instantiating the browser without arguments,
            # so it is running without a proxy. Stop it first.
            with pytest.raises(brozzler.ProxyError):
                worker.brozzle_page(browser, site, page)


def test_try_login(httpd):
    """Test try_login behavior."""
    response_urls = []

    def on_response(msg):
        response_urls.append(msg["params"]["response"]["url"])

    chrome_exe = brozzler.suggest_default_chrome_exe()
    form_url = "http://localhost:%s/site11/form1.html" % httpd.server_port
    form_url_other = "http://localhost:%s/site11/form2.html" % httpd.server_port
    favicon_url = "http://localhost:%s/favicon.ico" % httpd.server_port
    login_url = "http://localhost:%s/login-action" % httpd.server_port
    # When username and password are defined and initial page has login form,
    # detect login form, submit login, and then return to the initial page.
    username = "user1@example.com"  # input type="email"
    password = "pass1"
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        browser.browse_page(
            form_url, username=username, password=password, on_response=on_response
        )
    assert len(response_urls) == 4
    assert response_urls[0] == form_url
    assert response_urls[1] == favicon_url
    assert response_urls[2] == login_url
    assert response_urls[3] == form_url

    # We are now supporting a different type of form, we'll test that here.
    response_urls = []
    username = "user1"
    password = "pass1"
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        browser.browse_page(
            form_url_other,
            username=username,
            password=password,
            on_response=on_response,
        )
    assert len(response_urls) == 4
    assert response_urls[0] == form_url_other
    assert response_urls[1] == favicon_url
    assert response_urls[2] == login_url
    assert response_urls[3] == form_url_other

    # When username and password are not defined, just load the initial page.
    response_urls = []
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        browser.browse_page(form_url, on_response=on_response)
    assert len(response_urls) == 2
    assert response_urls[0] == form_url
    assert response_urls[1] == favicon_url

    # when the page doesn't have a form with username/password, don't submit it
    response_urls = []
    form_without_login_url = (
        "http://localhost:%s/site11/form-no-login.html" % httpd.server_port
    )
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        browser.browse_page(
            form_without_login_url,
            username=username,
            password=password,
            on_response=on_response,
        )
    assert len(response_urls) == 2
    assert response_urls[0] == form_without_login_url
    assert response_urls[1] == favicon_url


def test_network_monitoring(httpd):
    chrome_exe = brozzler.suggest_default_chrome_exe()
    url = "http://localhost:%s/site12/" % httpd.server_port
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        # This page has no requests beyond the page itself.
        browser.navigate_to_page(url + "blank_page.html")

        assert len(browser.websock_thread.active_connections) == 0
        idle = time.time() - browser.websock_thread.last_network_activity
        # The page should have finished loading less than a second ago.
        assert 0 < idle < 1

        time.sleep(2)
        idle = time.time() - browser.websock_thread.last_network_activity
        assert 2 < idle < 3

        # This page will request an endpoint (which responds after 3 seconds) after 1s.
        browser.navigate_to_page(url + "with_request.html")

        assert len(browser.websock_thread.active_connections) == 0
        idle = time.time() - browser.websock_thread.last_network_activity
        assert 0 < idle < 1

        time.sleep(1)
        # The request should have started by now.
        assert len(browser.websock_thread.active_connections) == 1

        time.sleep(4)
        # The request should have finished by now.
        assert len(browser.websock_thread.active_connections) == 0

        browser.navigate_to_page(url + "blank_page.html")
        time.sleep(2)
        # This should be considered idle already.
        start = time.time()
        browser._wait_for_idle(2, 10)
        elapsed = time.time() - start
        assert 0 < elapsed < 1

        # Now make sure _wait_for_idle does the right thing.
        browser.navigate_to_page(url + "with_request.html")
        time.sleep(1)
        start = time.time()
        browser._wait_for_idle(2, 10)
        # 3 second request + 2 seconds idle = 5 seconds
        elapsed = time.time() - start
        assert 5 < elapsed < 6


def test_chrome_error(httpd):
    chrome_exe = brozzler.suggest_default_chrome_exe()
    url = "http://nonexistent.invalid"
    worker = brozzler.BrozzlerWorker(None)
    site = brozzler.Site(None, {})
    page = brozzler.Page(None, {"url": url})
    with brozzler.Browser(chrome_exe=chrome_exe) as browser:
        with pytest.raises(brozzler.PageConnectionError):
            worker.brozzle_page(browser, site, page)
        assert page.redirect_url and page.redirect_url.startswith("chrome-error:")
        page.url = "http://localhost:%s/site1/" % httpd.server_port
        worker.brozzle_page(browser, site, page)
        assert not (page.redirect_url and page.redirect_url.startswith("chrome-error:"))



#!/usr/bin/env python
"""
brozzler/cli.py - brozzler command line executables

Copyright (C) 2014-2025 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import argparse
import base64
import datetime
import json
import logging
import os
import signal
import string
import sys
import threading
import traceback
import warnings

import doublethink
import requests
import rethinkdb as rdb
import structlog
import yaml

import brozzler
import brozzler.worker
from brozzler import suggest_default_chrome_exe
from brozzler.model import VideoCaptureOptions

r = rdb.RethinkDB()

logger = structlog.get_logger(logger_name=__name__)


def add_common_options(arg_parser, argv=None):
    argv = argv or sys.argv
    arg_parser.add_argument(
        "-q",
        "--quiet",
        dest="log_level",
        action="store_const",
        default=logging.INFO,
        const=logging.WARN,
        help="quiet logging",
    )
    arg_parser.add_argument(
        "-v",
        "--verbose",
        dest="log_level",
        action="store_const",
        default=logging.INFO,
        const=logging.DEBUG,
        help=("verbose logging"),
    )
    arg_parser.add_argument(
        "--trace",
        dest="log_level",
        action="store_const",
        default=logging.INFO,
        const=logging.DEBUG,
        help=("very verbose logging"),
    )
    arg_parser.add_argument(
        "--syslogd-log-prefix",
        dest="syslogd_log_prefix",
        action="store_true",
        help="add syslogd log level prefix for journalctl filtering",
    )
    arg_parser.add_argument(
        "--worker-id",
        dest="worker_id",
        help="ID for this worker, displayed in logs if provided",
    )
    # arg_parser.add_argument(
    #         '-s', '--silent', dest='log_level', action='store_const',
    #         default=logging.INFO, const=logging.CRITICAL)
    arg_parser.add_argument(
        "--version",
        action="version",
        version="brozzler %s - %s" % (brozzler.__version__, os.path.basename(argv[0])),
    )


def add_rethinkdb_options(arg_parser):
    arg_parser.add_argument(
        "--rethinkdb-servers",
        dest="rethinkdb_servers",
        default=os.environ.get("BROZZLER_RETHINKDB_SERVERS", "localhost"),
        help=(
            "rethinkdb servers, e.g. "
            "db0.foo.org,db0.foo.org:38015,db1.foo.org (default is the "
            "value of environment variable BROZZLER_RETHINKDB_SERVERS)"
        ),
    )
    arg_parser.add_argument(
        "--rethinkdb-db",
        dest="rethinkdb_db",
        default=os.environ.get("BROZZLER_RETHINKDB_DB", "brozzler"),
        help=(
            "rethinkdb database name (default is the value of environment "
            "variable BROZZLER_RETHINKDB_DB)"
        ),
    )


def rethinker(args):
    servers = args.rethinkdb_servers or "localhost"
    db = args.rethinkdb_db or os.environ.get("BROZZLER_RETHINKDB_DB") or "brozzler"
    return doublethink.Rethinker(servers.split(","), db)


def decorate_logger_name(a, b, event_dict):
    """Decorates the logger name with call location, if provided"""

    old_name = event_dict.get("logger_name")
    if old_name is None:
        return event_dict

    try:
        filename = event_dict.pop("filename")
        func_name = event_dict.pop("func_name")
        lineno = event_dict.pop("lineno")
    except KeyError:
        return event_dict
    new_name = f"{old_name}.{func_name}({filename}:{lineno})"
    event_dict["logger_name"] = new_name

    return event_dict


# https://manpages.debian.org/testing/libsystemd-dev/sd-daemon.3.en.html
def _systemd_log_prefix(_, level, log):
    SYSLOG_MAP = {
        "critical": 2,
        "error": 3,
        "exception": 3,
        "warn": 3,
        "warning": 3,
        "info": 6,
        "debug": 7,
        "notset": 7,
    }
    prefix = SYSLOG_MAP.get(level)
    if prefix is not None:
        log = f"<{prefix}>{log}"

    return log


def configure_logging(args):
    processors = [
        structlog.contextvars.merge_contextvars,
        structlog.processors.add_log_level,
        structlog.processors.StackInfoRenderer(),
        structlog.dev.set_exc_info,
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M:%S", utc=True),
        structlog.processors.CallsiteParameterAdder(
            [
                structlog.processors.CallsiteParameter.FILENAME,
                structlog.processors.CallsiteParameter.FUNC_NAME,
                structlog.processors.CallsiteParameter.LINENO,
                structlog.processors.CallsiteParameter.PROCESS,
                structlog.processors.CallsiteParameter.THREAD_NAME,
            ],
        ),
        decorate_logger_name,
        structlog.dev.ConsoleRenderer(),
    ]

    if args.syslogd_log_prefix:
        processors.append(_systemd_log_prefix)

    structlog.configure(
        processors=processors,
        wrapper_class=structlog.make_filtering_bound_logger(args.log_level),
        context_class=dict,
        logger_factory=structlog.PrintLoggerFactory(),
        cache_logger_on_first_use=False,
    )

    # Adds the worker ID to the global binding, if supplied
    if args.worker_id is not None:
        structlog.contextvars.bind_contextvars(worker_id=args.worker_id)

    # We still configure logging for now because its handlers
    # are used for the gunicorn spawned by the brozzler dashboard.
    logging.basicConfig(
        stream=sys.stderr,
        level=args.log_level,
        format=(
            "%(asctime)s %(process)d %(levelname)s %(threadName)s "
            "%(name)s.%(funcName)s(%(filename)s:%(lineno)d) %(message)s"
        ),
    )
    logging.getLogger("requests.packages.urllib3").setLevel(logging.WARN)
    warnings.simplefilter(
        "ignore", category=requests.packages.urllib3.exceptions.InsecureRequestWarning
    )
    warnings.simplefilter(
        "ignore", category=requests.packages.urllib3.exceptions.InsecurePlatformWarning
    )


class BetterArgumentDefaultsHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    """
    Like argparse.ArgumentDefaultsHelpFormatter but omits the default value
    for arguments with action='store_const'.
    """

    def _get_help_string(self, action):
        if isinstance(action, argparse._StoreConstAction):
            return action.help
        else:
            return super()._get_help_string(action)


def brozzle_page(argv=None):
    """
    Command line utility entry point for brozzling a single page. Opens url in
    a browser, running some javascript behaviors, and prints outlinks.
    """
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        description="brozzle-page - brozzle a single page",
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument("url", metavar="URL", help="page url")
    arg_parser.add_argument(
        "-e",
        "--chrome-exe",
        dest="chrome_exe",
        default=suggest_default_chrome_exe(),
        help="executable to use to invoke chrome",
    )
    arg_parser.add_argument(
        "--behavior-parameters",
        dest="behavior_parameters",
        default=None,
        help=(
            "json blob of parameters to populate the javascript behavior "
            'template, e.g. {"parameter_username":"x",'
            '"parameter_password":"y"}'
        ),
    )
    arg_parser.add_argument(
        "--username",
        dest="username",
        default=None,
        help="use this username to try to log in if a login form is found",
    )
    arg_parser.add_argument(
        "--password",
        dest="password",
        default=None,
        help="use this password to try to log in if a login form is found",
    )
    arg_parser.add_argument("--proxy", dest="proxy", default=None, help="http proxy")
    arg_parser.add_argument(
        "--no-headless",
        dest="headless",
        action="store_false",
        help="Do not run Chrome headlessly",
    )
    arg_parser.add_argument(
        "--browser_throughput",
        type=int,
        dest="download_throughput",
        default=-1,
        help="Chrome DevTools downloadThroughput for Network.emulateNetworkConditions",
    )
    arg_parser.add_argument(
        "--browser_window_height",
        type=int,
        dest="window_height",
        default=900,
        help="browser window height in pixels",
    )
    arg_parser.add_argument(
        "--browser_window_width",
        type=int,
        dest="window_width",
        default=1400,
        help="browser window width in pixels",
    )
    arg_parser.add_argument(
        "--stealth",
        dest="stealth",
        action="store_true",
        help="Try to avoid web bot detection",
    )
    arg_parser.add_argument(
        "--metrics_port",
        type=int,
        dest="metrics_port",
        default=0,
        help="Port for brozzler's Prometheus scrape endpoint",
    )
    arg_parser.add_argument(
        "--registry_url",
        dest="registry_url",
        default=None,
        help="http-sd-registry url, for Prometheus metrics discovery",
    )
    arg_parser.add_argument(
        "--env",
        dest="env",
        default=None,
        help="deployment environment for this brozzler instance, e.g., prod or qa",
    )
    arg_parser.add_argument(
        "--screenshot-full-page", dest="screenshot_full_page", action="store_true"
    )
    arg_parser.add_argument(
        "--skip-extract-outlinks", dest="skip_extract_outlinks", action="store_true"
    )
    arg_parser.add_argument(
        "--skip-visit-hashtags", dest="skip_visit_hashtags", action="store_true"
    )
    arg_parser.add_argument(
        "--skip-youtube-dl", dest="skip_youtube_dl", action="store_true"
    )
    arg_parser.add_argument(
        "--ytdlp_tmpdir",
        dest="ytdlp_tmpdir",
        default="/tmp",
        help="specify a temp dir for ytdlp; defaults to /tmp",
    )
    arg_parser.add_argument("--simpler404", dest="simpler404", action="store_true")
    add_common_options(arg_parser, argv)

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)
    brozzler.chrome.check_version(args.chrome_exe)

    behavior_parameters = {}
    if args.behavior_parameters:
        behavior_parameters = json.loads(args.behavior_parameters)
    site = brozzler.Site(
        None,
        {
            "id": -1,
            "seed": args.url,
            "behavior_parameters": behavior_parameters,
            "username": args.username,
            "password": args.password,
        },
    )
    page = brozzler.Page(None, {"url": args.url, "site_id": site.id})
    worker = brozzler.BrozzlerWorker(
        frontier=None,
        proxy=args.proxy,
        headless=args.headless,
        skip_extract_outlinks=args.skip_extract_outlinks,
        skip_visit_hashtags=args.skip_visit_hashtags,
        skip_youtube_dl=args.skip_youtube_dl,
        ytdlp_tmpdir=args.ytdlp_tmpdir,
        simpler404=args.simpler404,
        screenshot_full_page=args.screenshot_full_page,
        download_throughput=args.download_throughput,
        window_height=args.window_height,
        window_width=args.window_width,
        stealth=args.stealth,
        metrics_port=args.metrics_port,
        registry_url=args.registry_url,
        env=args.env,
        worker_id=args.worker_id,
    )

    def on_screenshot(screenshot_jpeg):
        OK_CHARS = string.ascii_letters + string.digits
        filename = "/tmp/{}-{:%Y%m%d%H%M%S}.jpg".format(
            "".join(ch if ch in OK_CHARS else "_" for ch in args.url),
            datetime.datetime.now(),
        )
        with open(filename, "wb") as f:
            f.write(screenshot_jpeg)
        logger.info("wrote screenshot", filename=filename)

    browser = brozzler.Browser(chrome_exe=args.chrome_exe)
    try:
        browser.start(
            proxy=args.proxy,
            window_height=args.window_height,
            window_width=args.window_width,
            headless=args.headless,
        )
        outlinks = worker.brozzle_page(
            browser,
            site,
            page,
            on_screenshot=on_screenshot,
            enable_youtube_dl=not worker._skip_youtube_dl,
        )
        logger.info("outlinks", outlinks=sorted(outlinks))
    except brozzler.ReachedLimit:
        logger.exception("reached limit")
    except brozzler.PageInterstitialShown:
        logger.exception("page interstitial shown")
    except brozzler.PageConnectionError:
        logger.exception("Error between Chrome and Warcprox")
    finally:
        browser.stop()


def brozzler_new_job(argv=None):
    """
    Command line utility entry point for queuing a new brozzler job. Takes a
    yaml brozzler job configuration file, creates job, sites, and pages objects
    in rethinkdb, which brozzler-workers will look at and start crawling.
    """
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        description="brozzler-new-job - queue new job with brozzler",
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument(
        "job_conf_file",
        metavar="JOB_CONF_FILE",
        help="brozzler job configuration file in yaml",
    )
    add_rethinkdb_options(arg_parser)
    add_common_options(arg_parser, argv)

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)

    rr = rethinker(args)
    frontier = brozzler.RethinkDbFrontier(rr)
    try:
        brozzler.new_job_file(frontier, args.job_conf_file)
    except brozzler.InvalidJobConf as e:
        print(
            "brozzler-new-job: invalid job file:", args.job_conf_file, file=sys.stderr
        )
        print(
            "  " + yaml.dump(e.errors).rstrip().replace("\n", "\n  "), file=sys.stderr
        )
        sys.exit(1)


def brozzler_new_site(argv=None):
    """
    Command line utility entry point for queuing a new brozzler site.
    Takes a seed url and creates a site and page object in rethinkdb, which
    brozzler-workers will look at and start crawling.
    """
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        description="brozzler-new-site - register site to brozzle",
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument("seed", metavar="SEED", help="seed url")
    add_rethinkdb_options(arg_parser)
    arg_parser.add_argument(
        "--time-limit",
        dest="time_limit",
        default=None,
        help="time limit in seconds for this site",
    )
    arg_parser.add_argument(
        "--ignore-robots",
        dest="ignore_robots",
        action="store_true",
        help="ignore robots.txt for this site",
    )
    arg_parser.add_argument(
        "--warcprox-meta",
        dest="warcprox_meta",
        help=(
            "Warcprox-Meta http request header to send with each request; "
            "must be a json blob, ignored unless warcprox features are "
            "enabled"
        ),
    )
    arg_parser.add_argument(
        "--behavior-parameters",
        dest="behavior_parameters",
        default=None,
        help=(
            "json blob of parameters to populate the javascript behavior "
            'template, e.g. {"parameter_username":"x",'
            '"parameter_password":"y"}'
        ),
    )
    arg_parser.add_argument(
        "--username",
        dest="username",
        default=None,
        help="use this username to try to log in if a login form is found",
    )
    arg_parser.add_argument(
        "--password",
        dest="password",
        default=None,
        help="use this password to try to log in if a login form is found",
    )
    arg_parser.add_argument(
        "--disable-video-capture",
        dest="disable_video",
        action="store_true",
        help="disable video capture for this site",
    )
    add_common_options(arg_parser, argv)

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)

    if args.disable_video:
        video_capture = VideoCaptureOptions.DISABLE_VIDEO_CAPTURE.value
    else:
        video_capture = VideoCaptureOptions.ENABLE_VIDEO_CAPTURE.value

    rr = rethinker(args)
    site = brozzler.Site(
        rr,
        {
            "seed": args.seed,
            "time_limit": int(args.time_limit) if args.time_limit else None,
            "ignore_robots": args.ignore_robots,
            "warcprox_meta": (
                json.loads(args.warcprox_meta) if args.warcprox_meta else None
            ),
            "behavior_parameters": (
                json.loads(args.behavior_parameters)
                if args.behavior_parameters
                else None
            ),
            "username": args.username,
            "password": args.password,
            "video_capture": video_capture,
        },
    )

    frontier = brozzler.RethinkDbFrontier(rr)
    brozzler.new_site(frontier, site)


def brozzler_worker(argv=None):
    """
    Main entry point for brozzler, gets sites and pages to brozzle from
    rethinkdb, brozzles them.
    """
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    add_rethinkdb_options(arg_parser)
    arg_parser.add_argument(
        "-e",
        "--chrome-exe",
        dest="chrome_exe",
        default=suggest_default_chrome_exe(),
        help="executable to use to invoke chrome",
    )
    arg_parser.add_argument(
        "-n",
        "--max-browsers",
        dest="max_browsers",
        default="1",
        help="max number of chrome instances simultaneously browsing pages",
    )
    arg_parser.add_argument("--proxy", dest="proxy", default=None, help="http proxy")
    arg_parser.add_argument(
        "--no-headless",
        dest="headless",
        action="store_false",
        help="Do not run Chrome headlessly",
    )
    arg_parser.add_argument(
        "--browser_throughput",
        type=int,
        dest="download_throughput",
        default=-1,
        help="Chrome DevTools downloadThroughput for Network.emulateNetworkConditions",
    )
    arg_parser.add_argument(
        "--browser_window_height",
        type=int,
        dest="window_height",
        default=900,
        help="browser window height in pixels",
    )
    arg_parser.add_argument(
        "--browser_window_width",
        type=int,
        dest="window_width",
        default=1400,
        help="browser window width in pixels",
    )
    arg_parser.add_argument(
        "--warcprox-auto",
        dest="warcprox_auto",
        action="store_true",
        help=(
            "when needed, choose an available instance of warcprox from "
            "the rethinkdb service registry"
        ),
    )
    arg_parser.add_argument(
        "--skip-extract-outlinks",
        dest="skip_extract_outlinks",
        action="store_true",
        help=argparse.SUPPRESS,
    )
    arg_parser.add_argument(
        "--skip-visit-hashtags",
        dest="skip_visit_hashtags",
        action="store_true",
        help=argparse.SUPPRESS,
    )
    arg_parser.add_argument(
        "--skip-youtube-dl",
        dest="skip_youtube_dl",
        action="store_true",
        help=argparse.SUPPRESS,
    )
    arg_parser.add_argument(
        "--ytdlp_tmpdir",
        dest="ytdlp_tmpdir",
        default="/tmp",
        help="argparse.SUPPRESS",
    )
    arg_parser.add_argument(
        "--ytdlp_proxy_file",
        dest="ytdlp_proxy_file",
        default="/opt/local/brozzler/ytdlp_proxy_endpoints.txt",
        help="argparse.SUPPRESS",
    )
    arg_parser.add_argument(
        "--stealth",
        dest="stealth",
        action="store_true",
        help="Try to avoid web bot detection",
    )
    arg_parser.add_argument(
        "--metrics_port",
        type=int,
        dest="metrics_port",
        default=0,
        help="Port for brozzler's Prometheus scrape endpoint",
    )
    arg_parser.add_argument(
        "--registry_url",
        dest="registry_url",
        default=None,
        help="http-sd-registry url, for Prometheus metrics discovery",
    )
    arg_parser.add_argument(
        "--env",
        dest="env",
        default=None,
        help="deployment environment for this brozzler instance, e.g., prod or qa",
    )
    add_common_options(arg_parser, argv)

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)
    brozzler.chrome.check_version(args.chrome_exe)

    def dump_state(signum, frame):
        signal.signal(signal.SIGQUIT, signal.SIG_IGN)
        try:
            state_strs = []
            frames = sys._current_frames()
            threads = {th.ident: th for th in threading.enumerate()}
            for ident in frames:
                if threads[ident]:
                    state_strs.append(str(threads[ident]))
                else:
                    state_strs.append("<???:thread:ident=%s>" % ident)
                stack = traceback.format_stack(frames[ident])
                state_strs.append("".join(stack))
            logger.info(
                "dumping state (caught signal)\n%s", signal=signum, state=state_strs
            )
        except BaseException:
            logger.exception("exception dumping state")
        finally:
            signal.signal(signal.SIGQUIT, dump_state)

    def get_ytdlp_proxy_endpoints():
        YTDLP_PROXY_ENDPOINTS_FILE = args.ytdlp_proxy_file
        try:
            # make list from file
            with open(YTDLP_PROXY_ENDPOINTS_FILE) as endpoints:
                ytdlp_proxy_endpoints = [line for line in endpoints.readlines()]
                if ytdlp_proxy_endpoints:
                    logger.info(
                        "running with ytdlp proxy endpoints file",
                        ytdlp_proxy_endpoints=YTDLP_PROXY_ENDPOINTS_FILE,
                    )
        except Exception:
            ytdlp_proxy_endpoints = []
            logger.info("running with empty proxy endpoints file")
        return ytdlp_proxy_endpoints

    rr = rethinker(args)
    frontier = brozzler.RethinkDbFrontier(rr)
    service_registry = doublethink.ServiceRegistry(rr)
    ytdlp_proxy_endpoints_from_file = get_ytdlp_proxy_endpoints()
    worker = brozzler.worker.BrozzlerWorker(
        frontier,
        service_registry,
        ytdlp_proxy_endpoints=ytdlp_proxy_endpoints_from_file,
        max_browsers=int(args.max_browsers),
        chrome_exe=args.chrome_exe,
        proxy=args.proxy,
        headless=args.headless,
        warcprox_auto=args.warcprox_auto,
        skip_extract_outlinks=args.skip_extract_outlinks,
        skip_visit_hashtags=args.skip_visit_hashtags,
        skip_youtube_dl=args.skip_youtube_dl,
        ytdlp_tmpdir=args.ytdlp_tmpdir,
        stealth=args.stealth,
        metrics_port=args.metrics_port,
        registry_url=args.registry_url,
        env=args.env,
        worker_id=args.worker_id,
    )

    signal.signal(signal.SIGQUIT, dump_state)
    signal.signal(signal.SIGTERM, lambda s, f: worker.stop())
    signal.signal(signal.SIGINT, lambda s, f: worker.stop())

    th = threading.Thread(target=worker.run, name="BrozzlerWorkerThread")
    th.start()
    th.join()
    logger.info("brozzler-worker is all done, exiting")


def brozzler_ensure_tables(argv=None):
    """
    Creates rethinkdb tables if they don't already exist. Brozzler
    (brozzler-worker, brozzler-new-job, etc) normally creates the tables it
    needs on demand at startup, but if multiple instances are starting up at
    the same time, you can end up with duplicate broken tables. So it's a good
    idea to use this utility at an early step when spinning up a cluster.
    """
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    add_rethinkdb_options(arg_parser)
    add_common_options(arg_parser, argv)

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)

    rr = rethinker(args)

    # services table
    doublethink.ServiceRegistry(rr)

    # sites, pages, jobs tables
    brozzler.frontier.RethinkDbFrontier(rr)


class Jsonner(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.isoformat()
        elif isinstance(o, bytes):
            return base64.b64encode(o).decode("ascii")
        else:
            return json.JSONEncoder.default(self, o)


def brozzler_list_jobs(argv=None):
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument(
        "--yaml",
        dest="yaml",
        action="store_true",
        help=("yaml output (default is json)"),
    )
    group = arg_parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--active", dest="active", action="store_true", help=("list active jobs")
    )
    group.add_argument("--all", dest="all", action="store_true", help=("list all jobs"))
    group.add_argument(
        "--job", dest="job", metavar="JOB_ID", help=("list only the specified job")
    )
    add_rethinkdb_options(arg_parser)
    add_common_options(arg_parser, argv)

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)

    rr = rethinker(args)
    if args.job is not None:
        try:
            job_id = int(args.job)
        except ValueError:
            job_id = args.job
        reql = rr.table("jobs").get(job_id)
        logger.debug("querying rethinkdb", query=reql)
        result = reql.run()
        if result:
            results = [reql.run()]
        else:
            logger.error("no such job with id", job_id=job_id)
            sys.exit(1)
    else:
        reql = rr.table("jobs").order_by("id")
        if args.active:
            reql = reql.filter({"status": "ACTIVE"})
        logger.debug("querying rethinkdb", query=reql)
        results = reql.run()
    if args.yaml:
        yaml.dump_all(
            results, stream=sys.stdout, explicit_start=True, default_flow_style=False
        )
    else:
        for result in results:
            print(json.dumps(result, cls=Jsonner, indent=2))


def brozzler_list_sites(argv=None):
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument(
        "--yaml",
        dest="yaml",
        action="store_true",
        help=("yaml output (default is json)"),
    )
    group = arg_parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--active", dest="active", action="store_true", help=("list all active sites")
    )
    group.add_argument(
        "--job", dest="job", metavar="JOB_ID", help=("list sites for a particular job")
    )
    group.add_argument(
        "--jobless",
        dest="jobless",
        action="store_true",
        help=("list all jobless sites"),
    )
    group.add_argument(
        "--site", dest="site", metavar="SITE_ID", help=("list only the specified site")
    )
    group.add_argument(
        "--all", dest="all", action="store_true", help=("list all sites")
    )
    add_rethinkdb_options(arg_parser)
    add_common_options(arg_parser, argv)

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)

    rr = rethinker(args)

    reql = rr.table("sites")
    if args.job:
        try:
            job_id = int(args.job)
        except ValueError:
            job_id = args.job
        reql = reql.get_all(job_id, index="job_id")
    elif args.jobless:
        reql = reql.filter(~r.row.has_fields("job_id"))
    elif args.active:
        reql = reql.between(
            ["ACTIVE", r.minval], ["ACTIVE", r.maxval], index="sites_last_disclaimed"
        )
    elif args.site:
        reql = reql.get_all(args.site)
    logger.debug("querying rethinkdb", query=reql)
    results = reql.run()
    if args.yaml:
        yaml.dump_all(
            results, stream=sys.stdout, explicit_start=True, default_flow_style=False
        )
    else:
        for result in results:
            print(json.dumps(result, cls=Jsonner, indent=2))


def brozzler_list_pages(argv=None):
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument(
        "--yaml",
        dest="yaml",
        action="store_true",
        help=("yaml output (default is json)"),
    )
    group = arg_parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--job",
        dest="job",
        metavar="JOB_ID",
        help=("list pages for all sites of a particular job"),
    )
    group.add_argument(
        "--site",
        dest="site",
        metavar="SITE_ID",
        help=("list pages for the specified site"),
    )
    # group.add_argument(
    #         '--page', dest='page', metavar='PAGE_ID', help=(
    #             'list only the specified page'))
    group = arg_parser.add_mutually_exclusive_group()
    group.add_argument(
        "--queued", dest="queued", action="store_true", help=("limit to queued pages")
    )
    group.add_argument(
        "--brozzled",
        dest="brozzled",
        action="store_true",
        help=("limit to pages that have already been brozzled"),
    )
    group.add_argument(
        "--claimed",
        dest="claimed",
        action="store_true",
        help=("limit to pages that are currently claimed by a brozzler worker"),
    )
    add_rethinkdb_options(arg_parser)
    add_common_options(arg_parser, argv)

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)

    rr = rethinker(args)
    if args.job:
        try:
            job_id = int(args.job)
        except ValueError:
            job_id = args.job
        reql = rr.table("sites").get_all(job_id, index="job_id")["id"]
        logger.debug("querying rethinkb", query=reql)
        site_ids = reql.run()
    elif args.site:
        try:
            site_ids = [int(args.site)]
        except ValueError:
            site_ids = [args.site]

    for site_id in site_ids:
        reql = rr.table("pages")
        if args.queued:
            reql = reql.between(
                [site_id, 0, r.minval], [site_id, 0, r.maxval], index="least_hops"
            )
        elif args.brozzled:
            reql = reql.between(
                [site_id, 1, r.minval],
                [site_id, r.maxval, r.maxval],
                index="least_hops",
            )
        else:
            reql = reql.between(
                [site_id, 0, r.minval],
                [site_id, r.maxval, r.maxval],
                index="least_hops",
            )
        reql = reql.order_by(index="least_hops")
        if args.claimed:
            reql = reql.filter({"claimed": True})
        logger.debug("querying rethinkb", query=reql)
        results = reql.run()
        if args.yaml:
            yaml.dump_all(
                results,
                stream=sys.stdout,
                explicit_start=True,
                default_flow_style=False,
            )
        else:
            for result in results:
                print(json.dumps(result, cls=Jsonner, indent=2))


def brozzler_purge(argv=None):
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        description="brozzler-purge - purge crawl state from rethinkdb",
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    group = arg_parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "--job",
        dest="job",
        metavar="JOB_ID",
        help=(
            "purge crawl state from rethinkdb for a job, including all sites and pages"
        ),
    )
    group.add_argument(
        "--site",
        dest="site",
        metavar="SITE_ID",
        help=("purge crawl state from rethinkdb for a site, including all pages"),
    )
    group.add_argument(
        "--finished-before",
        dest="finished_before",
        metavar="YYYY-MM-DD",
        help=(
            "purge crawl state from rethinkdb for a jobs that ended before this date"
        ),
    )
    arg_parser.add_argument(
        "--force",
        dest="force",
        action="store_true",
        help=("purge even if job or site is still has status ACTIVE"),
    )
    add_rethinkdb_options(arg_parser)
    add_common_options(arg_parser, argv)

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)

    rr = rethinker(args)
    brozzler.RethinkDbFrontier(rr)
    if args.job:
        try:
            job_id = int(args.job)
        except ValueError:
            job_id = args.job
        job = brozzler.Job.load(rr, job_id)
        if not job:
            logger.fatal("no such job", job_id=job_id)
            sys.exit(1)
        if job.status == "ACTIVE":
            if args.force:
                logger.warning(
                    "job has status ACTIVE, purging anyway because "
                    "--force was supplied",
                    job_id=job_id,
                )
            else:
                logger.fatal(
                    "refusing to purge job because status is ACTIVE "
                    "(override with --force)",
                    job_id=job_id,
                )
                sys.exit(1)
        _purge_job(rr, job_id)
    elif args.site:
        site_id = args.site
        site = brozzler.Site.load(rr, site_id)
        if not site:
            logger.fatal("no such job", job_id=job_id)
            sys.exit(1)
        if site.status == "ACTIVE":
            if args.force:
                logger.warning(
                    "site has status ACTIVE, purging anyway because "
                    "--force was supplied",
                    site_id=site_id,
                )
            else:
                logger.fatal(
                    "refusing to purge site because status is ACTIVE "
                    "(override with --force)",
                    site_id=site_id,
                )
                sys.exit(1)
        _purge_site(rr, site_id)
    elif args.finished_before:
        finished_before = datetime.datetime.strptime(
            args.finished_before, "%Y-%m-%d"
        ).replace(tzinfo=doublethink.UTC)
        reql = rr.table("jobs").filter(
            r.row["finished"]
            .default(r.maxval)
            .lt(finished_before)
            .or_(
                r.row["starts_and_stops"]
                .nth(-1)["stop"]
                .default(r.maxval)
                .lt(finished_before)
            )
        )
        logger.debug("retrieving jobs older than %s", finished_before, query=reql)
        for job in reql.run():
            # logging.info('job %s finished=%s starts_and_stops[-1]["stop"]=%s',
            #         job['id'], job.get('finished'),
            #         job.get('starts_and_stops', [{'stop':None}])[-1]['stop'])
            _purge_job(rr, job["id"])


def _purge_site(rr, site_id):
    reql = (
        rr.table("pages")
        .between(
            [site_id, r.minval, r.minval],
            [site_id, r.maxval, r.maxval],
            index="priority_by_site",
        )
        .delete()
    )
    site_logger = logger.bind(site_id=site_id)

    site_logger.debug("purging pages for site", query=reql)
    result = reql.run()
    site_logger.info("purged pages for site", result=result)

    reql = rr.table("sites").get(site_id).delete()
    site_logger.debug("purging site", query=reql)
    result = reql.run()
    site_logger.info("purged site", result=result)


def _purge_job(rr, job_id):
    job_logger = logger.bind(job_id=job_id)

    reql = rr.table("sites").get_all(job_id, index="job_id").get_field("id")
    job_logger.debug("querying rethinkdb", query=reql)
    site_ids = list(reql.run())
    for site_id in site_ids:
        _purge_site(rr, site_id)

    reql = rr.table("jobs").get(job_id).delete()
    job_logger.debug("purging job", query=reql)
    result = reql.run()
    job_logger.info("purged job", result=result)


def brozzler_list_captures(argv=None):
    """
    Handy utility for looking up entries in the rethinkdb "captures" table by
    url or sha1.
    """
    import urlcanon

    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument(
        "-p",
        "--prefix",
        dest="prefix",
        action="store_true",
        help=(
            "use prefix match for url (n.b. may not work as expected if "
            "searching key has query string because canonicalization can "
            "reorder query parameters)"
        ),
    )
    arg_parser.add_argument(
        "--yaml",
        dest="yaml",
        action="store_true",
        help=("yaml output (default is json)"),
    )
    add_rethinkdb_options(arg_parser)
    add_common_options(arg_parser, argv)
    arg_parser.add_argument(
        "url_or_sha1",
        metavar="URL_or_SHA1",
        help="url or sha1 to look up in captures table",
    )

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)

    rr = rethinker(args)

    if args.url_or_sha1[:5] == "sha1:":
        if args.prefix:
            logger.warning(
                "ignoring supplied --prefix option which does not apply "
                "to lookup by sha1"
            )
        # assumes it's already base32 (XXX could detect if hex and convert)
        sha1base32 = args.url_or_sha1[5:].upper()
        reql = rr.table("captures").between(
            [sha1base32, r.minval, r.minval],
            [sha1base32, r.maxval, r.maxval],
            index="sha1_warc_type",
        )
        logger.debug("querying rethinkdb", query=reql)
        results = reql.run()
    else:
        key = urlcanon.semantic(args.url_or_sha1).surt().decode("ascii")
        abbr_start_key = key[:150]
        if args.prefix:
            # surt is necessarily ascii and \x7f is the last ascii character
            abbr_end_key = key[:150] + "\x7f"
            end_key = key + "\x7f"
        else:
            abbr_end_key = key[:150]
            end_key = key
        reql = rr.table("captures").between(
            [abbr_start_key, r.minval],
            [abbr_end_key, r.maxval],
            index="abbr_canon_surt_timestamp",
            right_bound="closed",
        )
        reql = reql.order_by(index="abbr_canon_surt_timestamp")
        reql = reql.filter(
            lambda capture: (
                (capture["canon_surt"] >= key) & (capture["canon_surt"] <= end_key)
            )
        )
        logger.debug("querying rethinkdb", query=reql)
        results = reql.run()

    if args.yaml:
        yaml.dump_all(
            results, stream=sys.stdout, explicit_start=True, default_flow_style=False
        )
    else:
        for result in results:
            print(json.dumps(result, cls=Jsonner, indent=2))


def brozzler_stop_crawl(argv=None):
    argv = argv or sys.argv
    arg_parser = argparse.ArgumentParser(
        prog=os.path.basename(argv[0]),
        formatter_class=BetterArgumentDefaultsHelpFormatter,
    )
    group = arg_parser.add_mutually_exclusive_group(required=True)
    add_rethinkdb_options(arg_parser)
    group.add_argument(
        "--job",
        dest="job_id",
        metavar="JOB_ID",
        help=("request crawl stop for the specified job"),
    )
    group.add_argument(
        "--site",
        dest="site_id",
        metavar="SITE_ID",
        help=("request crawl stop for the specified site"),
    )
    add_common_options(arg_parser, argv)

    args = arg_parser.parse_args(args=argv[1:])
    configure_logging(args)

    rr = rethinker(args)
    if args.job_id:
        try:
            job_id = int(args.job_id)
        except ValueError:
            job_id = args.job_id
        job = brozzler.Job.load(rr, job_id)
        if not job:
            logger.fatal("job not found with", id=job_id)
            sys.exit(1)
        job.stop_requested = doublethink.utcnow()
        job.save()
    elif args.site_id:
        try:
            site_id = int(args.site_id)
        except ValueError:
            site_id = args.site_id
        site = brozzler.Site.load(rr, site_id)
        if not site:
            logger.fatal("site not found with", id=site_id)
            sys.exit(1)
        site.stop_requested = doublethink.utcnow()
        site.save()



#!/usr/bin/env python
"""
test_cli.py - test brozzler commands

Copyright (C) 2017 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import importlib.metadata
import importlib.util
import os
import subprocess

import doublethink
import pytest

import brozzler.cli


@pytest.fixture(scope="module")
def rethinker(request):
    db = request.param if hasattr(request, "param") else "ignoreme"
    servers = os.environ.get("BROZZLER_RETHINKDB_SERVERS", "localhost")
    return doublethink.Rethinker(db=db, servers=servers.split(","))


def console_scripts():
    # We do a dict comprehension here because the select filters aren't
    # available until Python 3.10's importlib.
    return {
        ep.name: ep
        for ep in importlib.metadata.distribution("brozzler").entry_points
        if ep.group == "console_scripts"
    }


def cli_commands():
    commands = set(console_scripts().keys())
    commands.remove("brozzler-wayback")
    if not importlib.util.find_spec("gunicorn"):
        commands.remove("brozzler-dashboard")

    if not importlib.util.find_spec("pywb"):
        commands.remove("brozzler-easy")

    return commands


@pytest.mark.parametrize("cmd", cli_commands())
def test_call_entrypoint(capsys, cmd):
    entrypoint = console_scripts()[cmd]
    callable = entrypoint.load()
    with pytest.raises(SystemExit):
        callable(["/whatever/bin/%s" % cmd, "--version"])
    out, err = capsys.readouterr()
    assert out == "brozzler %s - %s\n" % (brozzler.__version__, cmd)
    assert err == ""


@pytest.mark.parametrize("cmd", cli_commands())
def test_run_command(capsys, cmd):
    proc = subprocess.Popen(
        [cmd, "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    out, err = proc.communicate()
    # Remove lines from syntax warning in imported library
    filtered_lines = [
        line
        for line in err.decode("utf-8").splitlines()
        if "reppy" not in line and "re.compile" not in line
    ]
    assert filtered_lines == []
    assert out == ("brozzler %s - %s\n" % (brozzler.__version__, cmd)).encode("ascii")


@pytest.mark.parametrize("rethinker", ["rethinkdb"], indirect=True)  # build-in db
def test_rethinkdb_up(rethinker):
    """Check that rethinkdb is up and running."""
    # check that rethinkdb is listening and looks sane
    rr = rethinker
    tbls = rr.table_list().run()
    assert len(tbls) > 10


# XXX don't know why this test is failing in travis-ci and vagrant while
# test_call_entrypoint tests pass :( (also fails with capfd)
@pytest.mark.xfail
def test_stop_nonexistent_crawl(capsys):
    with pytest.raises(SystemExit):
        brozzler.cli.brozzler_stop_crawl(["brozzler-stop-crawl", "--site=123"])
    out, err = capsys.readouterr()
    assert err.endswith("site not found with id=123\n")
    assert out == ""

    with pytest.raises(SystemExit):
        brozzler.cli.brozzler_stop_crawl(["brozzler-stop-crawl", "--job=abc"])
    out, err = capsys.readouterr()
    assert err.endswith("""job not found with id='abc'\n""")
    assert out == ""



"""
brozzler/frontier.py - RethinkDbFrontier manages crawl jobs, sites and pages

Copyright (C) 2014-2018 Internet Archive

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import datetime
from typing import Dict, List

import doublethink
import rethinkdb as rdb
import structlog
import urlcanon

import brozzler

r = rdb.RethinkDB()


class UnexpectedDbResult(Exception):
    pass


def filter_claimable_site_ids(
    active_sites: List[Dict],
    reclaim_cooldown: int,
    max_sites_to_claim=1,
) -> List[str]:
    job_counts = {}
    claimable_sites = []
    now = datetime.datetime.now(datetime.timezone.utc)

    for site in active_sites:
        is_claimable = False

        # If site not claimed and not disclaimed within last 20 seconds
        if not site["claimed"] and site.get("last_disclaimed", 0) <= (
            now - datetime.timedelta(seconds=reclaim_cooldown)
        ):
            is_claimable = True

        # or site has been disclaimed more than an hour ago
        if "last_claimed" in site and site["last_claimed"] <= (
            now - datetime.timedelta(hours=1)
        ):
            is_claimable = True

        # Count number of claimed sites per job_id (optional field)
        if site["claimed"] and "max_claimed_sites" in site and "job_id" in site:
            job_id = site["job_id"]
            job_counts[job_id] = job_counts.get(job_id, 0) + 1

        if is_claimable:
            claimable_sites.append(site)

    site_ids_to_claim = []
    # gather sites that are under the max without going over
    for site in claimable_sites:
        if (
            "max_claimed_sites" in site
            and "job_id" in site
            and job_counts.get(site["job_id"], 0) < site["max_claimed_sites"]
        ):
            site_ids_to_claim.append(site["id"])
            job_counts[site["job_id"]] = job_counts.get(site["job_id"], 0) + 1

        if "max_claimed_sites" not in site or "job_id" not in site:
            site_ids_to_claim.append(site["id"])

        # short circuit if we already have more than requested
        if len(site_ids_to_claim) >= max_sites_to_claim:
            break

    return site_ids_to_claim


class RethinkDbFrontier:
    logger = structlog.get_logger(logger_name=__module__ + "." + __qualname__)

    def __init__(self, rr, shards=None, replicas=None):
        self.rr = rr
        self.shards = shards or len(rr.servers)
        self.replicas = replicas or min(len(rr.servers), 3)
        self._ensure_db()

    def _ensure_db(self):
        db_logger = self.logger.bind(dbname=self.rr.dbname)

        dbs = self.rr.db_list().run()
        if self.rr.dbname not in dbs:
            db_logger.info("creating rethinkdb database")
            self.rr.db_create(self.rr.dbname).run()
        tables = self.rr.table_list().run()
        if "sites" not in tables:
            db_logger.info("creating rethinkdb table 'sites' in database")
            self.rr.table_create(
                "sites", shards=self.shards, replicas=self.replicas
            ).run()
            self.rr.table("sites").index_create(
                "sites_last_disclaimed", [r.row["status"], r.row["last_disclaimed"]]
            ).run()
            self.rr.table("sites").index_create("job_id").run()
        if "pages" not in tables:
            db_logger.info("creating rethinkdb table 'pages' in database")
            self.rr.table_create(
                "pages", shards=self.shards, replicas=self.replicas
            ).run()
            self.rr.table("pages").index_create(
                "priority_by_site",
                [
                    r.row["site_id"],
                    r.row["brozzle_count"],
                    r.row["claimed"],
                    r.row["priority"],
                ],
            ).run()
            # this index is for displaying pages in a sensible order in the web
            # console
            self.rr.table("pages").index_create(
                "least_hops",
                [r.row["site_id"], r.row["brozzle_count"], r.row["hops_from_seed"]],
            ).run()
        if "jobs" not in tables:
            db_logger.info("creating rethinkdb table 'jobs' in database")
            self.rr.table_create(
                "jobs", shards=self.shards, replicas=self.replicas
            ).run()

    def _vet_result(self, result, **kwargs):
        # self.logger.debug("vetting expected=%s result=%s", kwargs, result)
        # {'replaced': 0, 'errors': 0, 'skipped': 0, 'inserted': 1, 'deleted': 0, 'generated_keys': ['292859c1-4926-4b27-9d87-b2c367667058'], 'unchanged': 0}
        for k in ["replaced", "errors", "skipped", "inserted", "deleted", "unchanged"]:
            if k in kwargs:
                expected = kwargs[k]
            else:
                expected = 0
            if isinstance(expected, list):
                if result.get(k) not in kwargs[k]:
                    raise UnexpectedDbResult(
                        "expected %r to be one of %r in %r" % (k, expected, result)
                    )
            else:
                if result.get(k) != expected:
                    raise UnexpectedDbResult(
                        "expected %r to be %r in %r" % (k, expected, result)
                    )

    def get_active_sites(self) -> List[Dict]:
        active_sites = (
            self.rr.table("sites", read_mode="majority")
            .between(
                ["ACTIVE", r.minval],
                ["ACTIVE", r.maxval],
                index="sites_last_disclaimed",
            )
            .pluck(
                "id",
                "last_disclaimed",
                "claimed",
                "last_claimed",
                "job_id",
                "max_claimed_sites",
            )
            .order_by(r.desc("claimed"), "last_disclaimed")
            .run()
        )
        return active_sites

    def claim_sites(self, n=1, reclaim_cooldown=20) -> List[Dict]:
        self.logger.debug("claiming up to %s sites to brozzle", n)

        active_sites = self.get_active_sites()
        site_ids_to_claim = filter_claimable_site_ids(
            active_sites, reclaim_cooldown, max_sites_to_claim=n
        )
        result = (
            self.rr.table("sites", read_mode="majority")
            .get_all(r.args(site_ids_to_claim))
            .update(  # mark the sites we're claiming, and return changed sites (our final claim
                # results)
                #
                # try to avoid a race condition resulting in multiple
                # brozzler-workers claiming the same site
                # see https://github.com/rethinkdb/rethinkdb/issues/3235#issuecomment-60283038
                r.branch(
                    r.or_(
                        r.row["claimed"].not_(),
                        r.row["last_claimed"].lt(r.now().sub(60 * 60)),
                    ),
                    {"claimed": True, "last_claimed": r.now()},
                    {},
                ),
                return_changes=True,
            )
        ).run()

        self._vet_result(
            result, replaced=list(range(n + 1)), unchanged=list(range(n + 1))
        )
        sites = []
        for i in range(result["replaced"]):
            if result["changes"][i]["old_val"]["claimed"]:
                self.logger.warning(
                    "re-claimed site that was still marked 'claimed' "
                    "because it was last claimed a long time ago, "
                    "and presumably some error stopped it from "
                    "being disclaimed",
                    last_claimed=result["changes"][i]["old_val"]["last_claimed"],
                )
            site = brozzler.Site(self.rr, result["changes"][i]["new_val"])
            sites.append(site)
        self.logger.debug("claimed %s sites", len(sites))
        if sites:
            return sites
        else:
            raise brozzler.NothingToClaim

    def enforce_time_limit(self, site):
        """
        Raises `brozzler.ReachedTimeLimit` if appropriate.
        """
        if site.time_limit and site.time_limit > 0 and site.elapsed() > site.time_limit:
            self.logger.debug(
                "site FINISHED_TIME_LIMIT!",
                time_limit=site.time_limit,
                elapsed=site.elapsed(),
                site=site,
            )
            raise brozzler.ReachedTimeLimit

    def claim_page(self, site, worker_id):
        # ignores the "claimed" field of the page, because only one
        # brozzler-worker can be working on a site at a time, and that would
        # have to be the worker calling this method, so if something is claimed
        # already, it must have been left that way because of some error
        result = (
            self.rr.table("pages")
            .between(
                [site.id, 0, r.minval, r.minval],
                [site.id, 0, r.maxval, r.maxval],
                index="priority_by_site",
            )
            .order_by(index=r.desc("priority_by_site"))
            .filter(
                lambda page: r.or_(
                    page.has_fields("retry_after").not_(), r.now() > page["retry_after"]
                )
            )
            .limit(1)
            .update(
                {"claimed": True, "last_claimed_by": worker_id}, return_changes="always"
            )
            .run()
        )
        self._vet_result(result, unchanged=[0, 1], replaced=[0, 1])
        if result["unchanged"] == 0 and result["replaced"] == 0:
            raise brozzler.NothingToClaim
        else:
            return brozzler.Page(self.rr, result["changes"][0]["new_val"])

    def has_outstanding_pages(self, site):
        results_iter = (
            self.rr.table("pages")
            .between(
                [site.id, 0, r.minval, r.minval],
                [site.id, 0, r.maxval, r.maxval],
                index="priority_by_site",
            )
            .limit(1)
            .run()
        )
        return len(list(results_iter)) > 0

    def completed_page(self, site, page):
        page.brozzle_count += 1
        page.claimed = False
        # XXX set priority?
        page.save()
        if page.redirect_url and page.hops_from_seed == 0:
            site.note_seed_redirect(page.redirect_url)
            site.save()

    def active_jobs(self):
        results = self.rr.table("jobs").filter({"status": "ACTIVE"}).run()
        for result in results:
            yield brozzler.Job(self.rr, result)

    def honor_stop_request(self, site):
        """Raises brozzler.CrawlStopped if stop has been requested."""
        site.refresh()
        if site.stop_requested and site.stop_requested <= doublethink.utcnow():
            self.logger.info("stop requested for site", site_id=site.id)
            raise brozzler.CrawlStopped

        if site.job_id:
            job = brozzler.Job.load(self.rr, site.job_id)
            if (
                job
                and job.stop_requested
                and job.stop_requested <= doublethink.utcnow()
            ):
                self.logger.info("stop requested for job", job_id=site.job_id)
                raise brozzler.CrawlStopped

    def _maybe_finish_job(self, job_id):
        """Returns True if job is finished."""
        job = brozzler.Job.load(self.rr, job_id)
        if not job:
            return False
        if job.status.startswith("FINISH"):
            self.logger.warning("%s is already %s", job, job.status)
            return True

        results = self.rr.table("sites").get_all(job_id, index="job_id").run()
        n = 0
        for result in results:
            site = brozzler.Site(self.rr, result)
            if not site.status.startswith("FINISH"):
                results.close()
                return False
            n += 1

        self.logger.info("all %s sites finished, job is FINISHED!", n, job_id=job.id)
        job.finish()
        job.save()
        return True

    def finished(self, site, status):
        self.logger.info("%s %s", status, site)
        site.status = status
        site.claimed = False
        site.last_disclaimed = doublethink.utcnow()
        site.starts_and_stops[-1]["stop"] = doublethink.utcnow()
        site.save()
        if site.job_id:
            self._maybe_finish_job(site.job_id)

    def disclaim_site(self, site, page=None):
        self.logger.info("disclaiming", site=site)
        site.claimed = False
        site.last_disclaimed = doublethink.utcnow()
        if not page and not self.has_outstanding_pages(site):
            self.finished(site, "FINISHED")
        else:
            site.save()
        if page:
            page.claimed = False
            page.save()

    def resume_job(self, job):
        job.status = "ACTIVE"
        job.stop_requested = None
        job.starts_and_stops.append({"start": doublethink.utcnow(), "stop": None})
        job.save()
        for site in self.job_sites(job.id):
            site.status = "ACTIVE"
            site.starts_and_stops.append({"start": doublethink.utcnow(), "stop": None})
            site.save()

    def resume_site(self, site):
        if site.job_id:
            # can't call resume_job since that would resume jobs's other sites
            job = brozzler.Job.load(self.rr, site.job_id)
            job.status = "ACTIVE"
            site.stop_requested = None
            job.starts_and_stops.append({"start": doublethink.utcnow(), "stop": None})
            job.save()
        site.status = "ACTIVE"
        site.starts_and_stops.append({"start": doublethink.utcnow(), "stop": None})
        site.save()

    def _build_fresh_page(self, site, parent_page, url, hops_off=0):
        url_for_crawling = urlcanon.whatwg(url)
        hashtag = (url_for_crawling.hash_sign + url_for_crawling.fragment).decode(
            "utf-8"
        )
        urlcanon.canon.remove_fragment(url_for_crawling)
        page = brozzler.Page(
            self.rr,
            {
                "url": str(url_for_crawling),
                "site_id": site.id,
                "job_id": site.job_id,
                "hops_from_seed": parent_page.hops_from_seed + 1,
                "hop_path": str(parent_page.hop_path if parent_page.hop_path else "")
                + "L",
                "via_page_id": parent_page.id,
                "via_page_url": parent_page.url,
                "hops_off_surt": hops_off,
                "hashtags": [hashtag] if hashtag else [],
            },
        )
        return page

    def _merge_page(self, existing_page, fresh_page):
        """
        Utility method for merging info from `brozzler.Page` instances
        representing the same url but with possibly different metadata.
        """
        existing_page.priority += fresh_page.priority
        existing_page.hashtags = list(
            set((existing_page.hashtags or []) + (fresh_page.hashtags or []))
        )
        existing_page.hops_off = min(existing_page.hops_off, fresh_page.hops_off)

    def _scope_and_enforce_robots(self, site, parent_page, outlinks):
        """
        Returns tuple (
            dict of {page_id: Page} of fresh `brozzler.Page` representing in
                scope links accepted by robots policy,
            set of in scope urls (canonicalized) blocked by robots policy,
            set of out-of-scope urls (canonicalized)).
        """
        pages = {}  # {page_id: Page, ...}
        blocked = set()
        out_of_scope = set()
        for url in outlinks or []:
            url_for_scoping = urlcanon.semantic(url)
            url_for_crawling = urlcanon.whatwg(url)
            decision = site.accept_reject_or_neither(
                url_for_scoping, parent_page=parent_page
            )
            if decision is True:
                hops_off = 0
            elif decision is None:
                decision = parent_page.hops_off < site.scope.get("max_hops_off", 0)
                hops_off = parent_page.hops_off + 1
            if decision is True:
                if brozzler.is_permitted_by_robots(site, str(url_for_crawling)):
                    fresh_page = self._build_fresh_page(
                        site, parent_page, url, hops_off
                    )
                    if fresh_page.id in pages:
                        self._merge_page(pages[fresh_page.id], fresh_page)
                    else:
                        pages[fresh_page.id] = fresh_page
                else:
                    blocked.add(str(url_for_crawling))
            else:
                out_of_scope.add(str(url_for_crawling))
        return pages, blocked, out_of_scope

    def scope_and_schedule_outlinks(self, site, parent_page, outlinks):
        decisions = {"accepted": set(), "blocked": set(), "rejected": set()}
        counts = {"added": 0, "updated": 0, "rejected": 0, "blocked": 0}

        fresh_pages, blocked, out_of_scope = self._scope_and_enforce_robots(
            site, parent_page, outlinks
        )
        decisions["blocked"] = blocked
        decisions["rejected"] = out_of_scope
        counts["blocked"] += len(blocked)
        counts["rejected"] += len(out_of_scope)

        # get existing pages from rethinkdb
        results = self.rr.table("pages").get_all(*fresh_pages.keys()).run()
        pages = {doc["id"]: brozzler.Page(self.rr, doc) for doc in results}

        # build list of pages to save, consisting of new pages, and existing
        # pages updated with higher priority and new hashtags
        for fresh_page in fresh_pages.values():
            decisions["accepted"].add(fresh_page.url)
            if fresh_page.id in pages:
                page = pages[fresh_page.id]
                self._merge_page(page, fresh_page)
                counts["updated"] += 1
            else:
                pages[fresh_page.id] = fresh_page
                counts["added"] += 1

        # make sure we're not stepping on our own toes in case we have a link
        # back to parent_page, which I think happens because of hashtags
        if parent_page.id in pages:
            self._merge_page(parent_page, pages[parent_page.id])
            del pages[parent_page.id]

        # insert/replace in batches of 50 to try to avoid this error:
        # "rethinkdb.errors.ReqlDriverError: Query size (167883036) greater than maximum (134217727) in:"
        # there can be many pages and each one can be very large (many videos,
        # in and out of scope links, etc)
        pages_list = list(pages.values())
        for batch in (pages_list[i : i + 50] for i in range(0, len(pages_list), 50)):
            try:
                self.logger.debug("inserting/replacing batch of %s pages", len(batch))
                reql = self.rr.table("pages").insert(batch, conflict="replace")
                self.logger.debug(
                    'running query self.rr.table("pages").insert(%r, '
                    'conflict="replace")',
                    batch,
                )
                reql.run()
            except Exception:
                self.logger.exception(
                    "problem inserting/replacing batch of %s pages",
                    len(batch),
                )

        parent_page.outlinks = {}
        for k in decisions:
            parent_page.outlinks[k] = list(decisions[k])
        parent_page.save()

        self.logger.info(
            "%s new links added, %s existing links updated, %s links "
            "rejected, %s links blocked by robots from %s",
            counts["added"],
            counts["updated"],
            counts["rejected"],
            counts["blocked"],
            parent_page,
        )

    def reached_limit(self, site, e):
        self.logger.info("reached_limit", site=site, e=e)
        assert isinstance(e, brozzler.ReachedLimit)
        if (
            site.reached_limit
            and site.reached_limit != e.warcprox_meta["reached-limit"]
        ):
            self.logger.warning(
                "reached limit %s but site had already reached limit %s",
                e.warcprox_meta["reached-limit"],
                self.reached_limit,
            )
        else:
            site.reached_limit = e.warcprox_meta["reached-limit"]
            self.finished(site, "FINISHED_REACHED_LIMIT")

    def job_sites(self, job_id):
        results = self.rr.table("sites").get_all(job_id, index="job_id").run()
        for result in results:
            yield brozzler.Site(self.rr, result)

    def seed_page(self, site_id):
        results = (
            self.rr.table("pages")
            .between(
                [site_id, r.minval, r.minval, r.minval],
                [site_id, r.maxval, r.maxval, r.maxval],
                index="priority_by_site",
            )
            .filter({"hops_from_seed": 0})
            .run()
        )
        pages = list(results)
        if len(pages) > 1:
            self.logger.warning("more than one seed page?", site_id=site_id)
        if len(pages) < 1:
            return None
        return brozzler.Page(self.rr, pages[0])

    def site_pages(self, site_id, brozzled=None):
        """
        Args:
            site_id (str or int):
            brozzled (bool): if true, results include only pages that have
                been brozzled at least once; if false, only pages that have
                not been brozzled; and if None (the default), all pages
        Returns:
            iterator of brozzler.Page
        """
        query = self.rr.table("pages").between(
            [site_id, 1 if brozzled is True else 0, r.minval, r.minval],
            [site_id, 0 if brozzled is False else r.maxval, r.maxval, r.maxval],
            index="priority_by_site",
        )
        self.logger.debug("running query", query=query)
        results = query.run()
        for result in results:
            self.logger.debug("yielding result", result=result)
            yield brozzler.Page(self.rr, result)
